# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for cloud computing provisioning backend functions."""

import datetime as dt
from collections.abc import Generator
from contextlib import contextmanager
from unittest import mock

from django.contrib.postgres.functions import TransactionNow
from django.db import connection
from django.utils import timezone

from debusine.db.context import context
from debusine.db.models import ScopeWorkerPool, Token, Worker, WorkerPool
from debusine.db.models.lookups import LookupCache
from debusine.db.models.worker_pools import (
    WorkerPoolStatistics,
    WorkerPoolTaskExecutionStatistics,
)
from debusine.db.permissioncontext import PermissionContext
from debusine.db.playground import scenarios
from debusine.server.collections.lookup import lookup_single
from debusine.server.provisioning import (
    ScopeWorkerAllocator,
    WorkerAllocator,
    create_new_workers,
    provision,
    scrub,
    scrub_workers,
    terminate_idle,
)
from debusine.server.worker_pools.models import WorkerPoolLimits
from debusine.test.django import TestCase


class ProvisionTests(TestCase):
    """Test :py:func:`provision`."""

    def test_no_idle(self) -> None:
        """Call create_new_workers if there are no idle workers."""
        with (
            mock.patch(
                "debusine.server.provisioning.terminate_idle",
                return_value=False,
            ) as ti,
            mock.patch(
                "debusine.server.provisioning.create_new_workers"
            ) as cnw,
        ):
            provision()
        ti.assert_called()
        cnw.assert_called()

    def test_has_idle(self) -> None:
        """No not call create_new_workers if there are idle workers."""
        with (
            mock.patch(
                "debusine.server.provisioning.terminate_idle", return_value=True
            ) as ti,
            mock.patch(
                "debusine.server.provisioning.create_new_workers"
            ) as cnw,
        ):
            provision()
        ti.assert_called()
        cnw.assert_not_called()

    def test_uses_local_context(self) -> None:
        """
        Provisioning uses a local context.

        Terminating workers may need to retry running work requests, which
        computes dynamic data, which may add entries to the lookup cache.
        Those entries do not persist between provisioning runs.
        """
        artifact, _ = self.playground.create_artifact()

        def fake_terminate_idle() -> None:
            lookup_single(
                f"{artifact.id}@artifacts",
                artifact.workspace,
                pc=PermissionContext.for_ui_user(
                    self.playground.get_default_user()
                ),
            )

        with (
            mock.patch(
                "debusine.server.provisioning.terminate_idle",
                side_effect=fake_terminate_idle,
            ),
            mock.patch("debusine.server.provisioning.create_new_workers"),
        ):
            provision()

        self.assertEqual(context.lookup_cache, LookupCache())


class TerminateIdleTests(TestCase):
    """Test :py:func:`terminate_idle`."""

    @contextmanager
    def mock_terminate(self) -> Generator[mock.MagicMock]:
        """Mock WorkerPool.terminate_idle."""
        with mock.patch.object(
            WorkerPool,
            "terminate_worker",
            autospec=True,
        ) as term:
            yield term

    def _create_worker(
        self, idle_minutes: int, worker_pool: WorkerPool | None = None
    ) -> Worker:
        worker = self.playground.create_worker(worker_pool=worker_pool)
        worker.registered_at = timezone.now() - dt.timedelta(
            seconds=idle_minutes * 60
        )
        worker.instance_created_at = timezone.now() - dt.timedelta(
            seconds=idle_minutes * 60
        )
        worker.save()
        return worker

    def test_no_workers(self) -> None:
        """Call with no workers at all."""
        with self.mock_terminate() as term:
            self.assertFalse(terminate_idle())
        term.assert_not_called()

    def test_no_idle(self) -> None:
        """Call with no idle workers."""
        worker_pool = self.playground.create_worker_pool("test")
        self._create_worker(idle_minutes=1, worker_pool=None)
        self._create_worker(idle_minutes=1, worker_pool=worker_pool)
        with self.mock_terminate() as term:
            self.assertFalse(terminate_idle(minutes=10))
        term.assert_not_called()

    def test_one_static_idle(self) -> None:
        """Call with one idle static worker."""
        worker_pool = self.playground.create_worker_pool("test")
        self._create_worker(idle_minutes=11, worker_pool=None)
        self._create_worker(idle_minutes=1, worker_pool=worker_pool)
        with self.mock_terminate() as term:
            self.assertFalse(terminate_idle(minutes=10))
        term.assert_not_called()

    def test_one_dynamic_idle(self) -> None:
        """Call with one idle cloud worker."""
        worker_pool = self.playground.create_worker_pool("test")
        self._create_worker(idle_minutes=1, worker_pool=None)
        cloud = self._create_worker(idle_minutes=11, worker_pool=worker_pool)
        with self.mock_terminate() as term:
            self.assertTrue(terminate_idle(minutes=10))
        term.assert_called_with(worker_pool, cloud)

    def test_pool_disabled(self) -> None:
        worker_pool = self.playground.create_worker_pool("test", enabled=False)
        self._create_worker(idle_minutes=11, worker_pool=worker_pool)
        with self.mock_terminate() as term:
            self.assertFalse(terminate_idle(minutes=10))
        term.assert_not_called()

    def test_tombstone(self) -> None:
        worker_pool = self.playground.create_worker_pool("test")
        worker = self._create_worker(idle_minutes=11, worker_pool=worker_pool)
        worker.instance_created_at = None
        worker.save()
        with self.mock_terminate() as term:
            self.assertFalse(terminate_idle(minutes=10))
        term.assert_not_called()


class CreateNewWorkersTests(TestCase):
    """Test :py:func:`create_new_workers`."""

    def test_noop(self) -> None:
        """Call the function: it does nothing at the moment."""
        create_new_workers()


class WorkerAllocatorTests(TestCase):
    """Test :py:class:`WorkerAllocator`."""

    def test_static_workers_count_no_workers(self) -> None:
        """Static workers count with no workers is 0."""
        allocator = WorkerAllocator()
        self.assertEqual(allocator.static_workers, {})

    def test_static_workers_count(self) -> None:
        """Static workers count counts only connected static workers."""
        pool = self.playground.create_worker_pool()
        for connected, in_pool, dyn_archs, static_archs in (
            (True, False, ["amd64"], None),
            (True, False, ["amd64"], ["i386"]),  # i386 overrides amd64
            (True, False, None, ["amd64"]),
            (True, False, ["ppc64el"], None),
            (False, False, ["amd64"], None),
            (True, True, ["amd64"], None),
        ):
            worker = self.playground.create_worker(
                worker_pool=pool if in_pool else None,
                extra_dynamic_metadata=(
                    {"system:architectures": dyn_archs} if dyn_archs else None
                ),
            )
            if static_archs:
                worker.static_metadata = {"system:architectures": static_archs}
                worker.save()
            if connected:
                worker.mark_connected()
        allocator = WorkerAllocator()
        self.assertEqual(
            allocator.static_workers,
            {
                "amd64": 2,
                "i386": 1,
                "ppc64el": 1,
            },
        )

    def test_start_of_month(self) -> None:
        """Ensure WorkerAllocator computes start_of_month correctly."""
        utc = dt.UTC
        for today, expected in (
            (
                dt.datetime(2025, 1, 2, 3, 4, 5, 6, tzinfo=utc),
                dt.datetime(2025, 1, 1, tzinfo=utc),
            ),
            (
                dt.datetime(2025, 1, 1, tzinfo=utc),
                dt.datetime(2025, 1, 1, tzinfo=utc),
            ),
        ):
            with self.subTest(today):
                with mock.patch(
                    "debusine.server.provisioning.timezone.now",
                    return_value=today,
                ):
                    allocator = WorkerAllocator()
                self.assertEqual(allocator.start_of_month, expected)

    def test_limit_by_instances_no_limits(self) -> None:
        """Call limit_by_max_active_instances with no limits."""
        pool = self.playground.create_worker_pool(
            limits=WorkerPoolLimits(max_active_instances=None)
        )
        worker_allocator = WorkerAllocator()
        for value in 0, 10, 1000:
            with self.subTest(value=value):
                self.assertEqual(
                    worker_allocator.limit_by_max_active_instances(pool, value),
                    value,
                )

    def test_limit_by_instances_no_workers(self) -> None:
        """Call limit_by_max_active_instances with no workers."""
        pool = self.playground.create_worker_pool(
            limits=WorkerPoolLimits(max_active_instances=5)
        )
        worker_allocator = WorkerAllocator()
        # 1 worker tombstone
        worker = self.playground.create_worker(worker_pool=pool)
        worker.instance_created_at = None
        worker.save()
        for value, expected in (
            (0, 0),
            (1, 1),
            (5, 5),
            (6, 5),
            (10, 5),
            (1000, 5),
        ):
            with self.subTest(value=value):
                self.assertEqual(
                    worker_allocator.limit_by_max_active_instances(pool, value),
                    expected,
                )

    def test_limit_by_instances(self) -> None:
        """Call limit_by_max_active_instances with workers."""
        pool = self.playground.create_worker_pool(
            limits=WorkerPoolLimits(max_active_instances=5)
        )
        worker_allocator = WorkerAllocator()
        for value, workers, expected in (
            (0, 1, 0),
            (1, 1, 1),
            (1, 5, 0),
            (1, 6, 0),
            (5, 0, 5),
            (5, 1, 4),
            (5, 5, 0),
            (6, 0, 5),
            (6, 1, 4),
            (6, 5, 0),
        ):
            Worker.objects.all().delete()
            Token.objects.filter(
                token_type__in={
                    Token.Types.WORKER_ACTIVATION,
                    Token.Types.WORKER,
                }
            ).delete()
            for _ in range(workers):
                self.playground.create_worker(worker_pool=pool)
            with self.subTest(value=value):
                self.assertEqual(
                    worker_allocator.limit_by_max_active_instances(pool, value),
                    expected,
                )

    def test_get_used_seconds_no_stats(self) -> None:
        """Call get_used_seconds with no WorkerPoolStatistics."""
        pool = self.playground.create_worker_pool()
        allocator = WorkerAllocator()
        self.assertEqual(allocator.get_used_seconds(pool), 0)

    def test_get_used_seconds_only_current_month(self) -> None:
        """get_used_seconds uses only data from current month."""
        pool = self.playground.create_worker_pool()
        worker = self.playground.create_worker(worker_pool=pool)
        allocator = WorkerAllocator()
        expected_used_seconds = 0

        for since_start_of_month, runtime, runtime_this_month in (
            (-dt.timedelta(days=60), 1, 0),
            (dt.timedelta(), 2, 0),
            (dt.timedelta(seconds=2), 10, 2),
            (dt.timedelta(days=60), 4, 4),
        ):
            ws = WorkerPoolStatistics.objects.create(
                worker_pool=pool, worker=worker, runtime=runtime
            )
            ws.timestamp = allocator.start_of_month + since_start_of_month
            ws.save()
            expected_used_seconds += runtime_this_month

        self.assertEqual(
            allocator.get_used_seconds(pool), expected_used_seconds
        )

    def test_get_used_seconds_adds_connected(self) -> None:
        """get_used_seconds adds runtime from connected workers."""
        pool = self.playground.create_worker_pool()
        workers = [
            self.playground.create_worker(worker_pool=pool) for _ in range(2)
        ]
        workers.append(
            self.playground.create_worker(
                worker_pool=self.playground.create_worker_pool(
                    name="test2", provider_account=pool.provider_account
                )
            )
        )
        allocator = WorkerAllocator()

        ws = WorkerPoolStatistics.objects.create(
            worker_pool=pool, worker=workers[0], runtime=10
        )
        ws.timestamp = allocator.start_of_month + dt.timedelta(seconds=2)
        ws.save()
        expected_used_seconds = 2

        workers[0].connected_at = workers[0].instance_created_at = (
            TransactionNow() - dt.timedelta(minutes=1)
        )
        workers[0].save()
        expected_used_seconds += 60

        workers[1].connected_at = workers[1].instance_created_at = (
            allocator.start_of_month - dt.timedelta(hours=1)
        )
        workers[1].save()
        with connection.cursor() as cursor:
            cursor.execute("SELECT CURRENT_TIMESTAMP")
            [now] = cursor.fetchone()
        expected_used_seconds += int(
            (now - allocator.start_of_month).total_seconds()
        )

        # This worker's runtime is not added, since it's in a different pool.
        workers[2].connected_at = workers[2].instance_created_at = (
            allocator.start_of_month
        )
        workers[2].save()

        self.assertEqual(
            allocator.get_used_seconds(pool), expected_used_seconds
        )

    def test_limit_max_seconds_per_month_no_limit(self) -> None:
        """Call with ScopeWorkerPool having None as limit: no limits."""
        pool = self.playground.create_worker_pool()
        allocator = WorkerAllocator()
        for val in (0, 10, 100, 1000):
            with self.subTest(val=val):
                self.assertEqual(
                    allocator.limit_by_pool_target_max_seconds_per_month(
                        pool, val
                    ),
                    val,
                )

    def test_limit_max_seconds_per_month(self) -> None:
        """Call with various limit scenarios."""
        pool = self.playground.create_worker_pool()
        allocator = WorkerAllocator()
        for limit, used, amount, expected in (
            # There are WorkerPoolStatistics not exceeding target
            # by more than TASK_DURATION_SECONDS
            (1000, 200, 1, 1),
            # There are WorkerPoolStatistics not exceeding target
            # by less than TASK_DURATION_SECONDS
            (1000, 500, 1, 0),
            # There are WorkerPoolStatistics exceeding target
            (1000, 1000, 1, 0),
            (1000, 1200, 1, 0),
            # Amount can get limited not just zeroed
            (6000, 0, 10, 10),
            (6000, 1, 10, 9),
            # Limit of 0 means no allocation allowed
            (0, 0, 1, 0),
        ):
            with (
                self.subTest(limit=limit, used=used, amount=amount),
                mock.patch.object(
                    allocator, "get_used_seconds", return_value=used
                ),
            ):
                pool.limits = {"target_max_seconds_per_month": limit}
                pool.save()
                self.assertEqual(
                    allocator.limit_by_pool_target_max_seconds_per_month(
                        pool, amount
                    ),
                    expected,
                )

    def test_process_scope(self) -> None:
        scope = self.playground.get_default_scope()
        allocator = WorkerAllocator()
        with mock.patch(
            "debusine.server.provisioning.ScopeWorkerAllocator.process_scope",
            autospec=True,
        ) as process_scope:
            allocator.process_scope(scope)

        process_scope.assert_called()
        scope_allocator = process_scope.call_args[0][0]
        self.assertIs(scope_allocator.worker_allocator, allocator)
        self.assertIs(scope_allocator.scope, scope)


class ScopeWorkerAllocatorTests(TestCase):
    """Test :py:class:`ScopeWorkerAllocator`."""

    scenario = scenarios.DefaultContext()

    def test_limit_max_seconds_per_month_no_scopeworkerpool(self) -> None:
        """Call with no ScopeWorkerPool: no limits."""
        pool = self.playground.create_worker_pool()
        worker_allocator = WorkerAllocator()
        allocator = ScopeWorkerAllocator(worker_allocator, self.scenario.scope)
        for val in (0, 10, 100, 1000):
            with self.subTest(val=val):
                self.assertEqual(
                    allocator.limit_by_scope_pool_target_max_seconds_per_month(
                        pool, val
                    ),
                    val,
                )

    def test_limit_max_seconds_per_month_no_limit(self) -> None:
        """Call with ScopeWorkerPool having None as limit: no limits."""
        pool = self.playground.create_worker_pool()
        ScopeWorkerPool.objects.create(
            worker_pool=pool,
            scope=self.scenario.scope,
            limits={"target_max_seconds_per_month": None},
        )
        worker_allocator = WorkerAllocator()
        allocator = ScopeWorkerAllocator(worker_allocator, self.scenario.scope)
        for val in (0, 10, 100, 1000):
            with self.subTest(val=val):
                self.assertEqual(
                    allocator.limit_by_scope_pool_target_max_seconds_per_month(
                        pool, val
                    ),
                    val,
                )

    def test_limit_max_seconds_per_month(self) -> None:
        """Call with various limit scenarios."""
        pool = self.playground.create_worker_pool()
        worker_allocator = WorkerAllocator()
        allocator = ScopeWorkerAllocator(worker_allocator, self.scenario.scope)
        for limit, used, amount, expected in (
            # There are WorkerPoolTaskExecutionStatistics not exceeding target
            # by more than TASK_DURATION_SECONDS
            (1000, 200, 1, 1),
            # There are WorkerPoolTaskExecutionStatistics not exceeding target
            # by less than TASK_DURATION_SECONDS
            (1000, 500, 1, 0),
            # There are WorkerPoolTaskExecutionStatistics exceeding target
            (1000, 1000, 1, 0),
            (1000, 1200, 1, 0),
            # Amount can get limited not just zeroed
            (6000, 0, 10, 10),
            (6000, 1, 10, 9),
            # Limit of 0 means no allocation allowed
            (0, 0, 1, 0),
        ):
            with (
                self.subTest(limit=limit, used=used, amount=amount),
                mock.patch.object(
                    allocator, "get_used_seconds", return_value=used
                ),
            ):
                ScopeWorkerPool.objects.all().delete()
                ScopeWorkerPool.objects.create(
                    worker_pool=pool,
                    scope=self.scenario.scope,
                    limits={"target_max_seconds_per_month": limit},
                )
                self.assertEqual(
                    allocator.limit_by_scope_pool_target_max_seconds_per_month(
                        pool, amount
                    ),
                    expected,
                )

    def test_get_used_seconds_no_stats(self) -> None:
        """Call get_used_seconds with no WorkerPoolTaskExecutionStatistics."""
        pool = self.playground.create_worker_pool()
        worker_allocator = WorkerAllocator()
        allocator = ScopeWorkerAllocator(worker_allocator, self.scenario.scope)
        self.assertEqual(allocator.get_used_seconds(pool), 0)

    def test_get_used_seconds_only_current_month(self) -> None:
        """get_used_seconds uses only data from current month."""
        pool = self.playground.create_worker_pool()
        worker = self.playground.create_worker(worker_pool=pool)
        worker_allocator = WorkerAllocator()
        allocator = ScopeWorkerAllocator(worker_allocator, self.scenario.scope)
        expected_used_seconds = 0

        for since_start_of_month, runtime, runtime_this_month in (
            (-dt.timedelta(days=60), 1, 0),
            (dt.timedelta(), 2, 0),
            (dt.timedelta(seconds=2), 10, 2),
            (dt.timedelta(days=60), 4, 4),
        ):
            ws = WorkerPoolTaskExecutionStatistics.objects.create(
                worker_pool=pool,
                worker=worker,
                scope=self.scenario.scope,
                runtime=runtime,
            )
            ws.timestamp = (
                worker_allocator.start_of_month + since_start_of_month
            )
            ws.save()
            expected_used_seconds += runtime_this_month

        self.assertEqual(
            allocator.get_used_seconds(pool), expected_used_seconds
        )

    def test_get_used_seconds_only_selected_scope_pool(self) -> None:
        """get_used_seconds uses only data from the selected scope and pool."""
        worker_allocator = WorkerAllocator()

        provider_account = self.playground.create_cloud_provider_account_asset()

        scope1 = self.scenario.scope
        scope2 = self.playground.get_or_create_scope(name="scope2")

        pool1 = self.playground.create_worker_pool(
            "pool1", provider_account=provider_account
        )
        pool2 = self.playground.create_worker_pool(
            "pool2", provider_account=provider_account
        )

        # A static worker should not interfere
        self.playground.create_worker()
        w1 = self.playground.create_worker(worker_pool=pool1)
        w2 = self.playground.create_worker(worker_pool=pool2)

        # Work done in pool1 for scope1
        WorkerPoolTaskExecutionStatistics.objects.create(
            worker_pool=pool1,
            worker=w1,
            scope=scope1,
            timestamp=worker_allocator.start_of_month,
            runtime=1,
        )
        # Work done in pool2 for scope1
        WorkerPoolTaskExecutionStatistics.objects.create(
            worker_pool=pool2,
            worker=w2,
            scope=scope1,
            timestamp=worker_allocator.start_of_month,
            runtime=2,
        )
        # Work done in pool1 for scope2
        WorkerPoolTaskExecutionStatistics.objects.create(
            worker_pool=pool1,
            worker=w1,
            scope=scope2,
            timestamp=worker_allocator.start_of_month,
            runtime=4,
        )
        # Work done in pool2 for scope2
        WorkerPoolTaskExecutionStatistics.objects.create(
            worker_pool=pool2,
            worker=w2,
            scope=scope2,
            timestamp=worker_allocator.start_of_month,
            runtime=8,
        )

        allocator1 = ScopeWorkerAllocator(worker_allocator, scope1)
        self.assertEqual(allocator1.get_used_seconds(pool1), 1)
        self.assertEqual(allocator1.get_used_seconds(pool2), 2)

        allocator2 = ScopeWorkerAllocator(worker_allocator, scope2)
        self.assertEqual(allocator2.get_used_seconds(pool1), 4)
        self.assertEqual(allocator2.get_used_seconds(pool2), 8)

    def test_get_used_seconds_adds_running(self) -> None:
        """get_used_seconds adds runtime from running work requests."""
        pool = self.playground.create_worker_pool()
        workers = [
            self.playground.create_worker(worker_pool=pool) for _ in range(3)
        ]
        workers.append(
            self.playground.create_worker(
                worker_pool=self.playground.create_worker_pool(
                    name="test2", provider_account=pool.provider_account
                )
            )
        )
        worker_allocator = WorkerAllocator()
        allocator = ScopeWorkerAllocator(worker_allocator, self.scenario.scope)

        ws = WorkerPoolTaskExecutionStatistics.objects.create(
            worker_pool=pool,
            worker=workers[0],
            scope=self.scenario.scope,
            runtime=10,
        )
        ws.timestamp = worker_allocator.start_of_month + dt.timedelta(seconds=2)
        ws.save()
        expected_used_seconds = 2

        workspace_in_other_scope = self.playground.create_workspace(
            scope=self.playground.get_or_create_scope(name="other-scope")
        )
        work_requests = [
            self.playground.create_worker_task(workspace=workspace)
            for workspace in (
                self.scenario.workspace,
                self.scenario.workspace,
                workspace_in_other_scope,
                self.scenario.workspace,
            )
        ]
        for worker, work_request in zip(workers, work_requests):
            self.playground.advance_work_request(
                work_request, mark_running=True, worker=worker
            )

        work_requests[0].started_at = TransactionNow() - dt.timedelta(minutes=1)
        work_requests[0].save()
        expected_used_seconds += 60

        work_requests[1].started_at = (
            worker_allocator.start_of_month - dt.timedelta(hours=1)
        )
        work_requests[1].save()
        with connection.cursor() as cursor:
            cursor.execute("SELECT CURRENT_TIMESTAMP")
            [now] = cursor.fetchone()
        expected_used_seconds += int(
            (now - worker_allocator.start_of_month).total_seconds()
        )

        # This work request's runtime is not added, since it's in a
        # different scope.
        work_requests[2].started_at = worker_allocator.start_of_month
        work_requests[2].save()

        # This work request's runtime is not added, since it's in a
        # different pool.
        work_requests[3].started_at = worker_allocator.start_of_month
        work_requests[3].save()

        self.assertEqual(
            allocator.get_used_seconds(pool), expected_used_seconds
        )

    def test_allocate_workers(self) -> None:
        pool = WorkerPool()
        worker_allocator = WorkerAllocator()
        allocator = ScopeWorkerAllocator(worker_allocator, self.scenario.scope)

        asked = [10, 10, 10]
        limited_to = [0, 1, 10]
        expected = [0, 1, 10]
        with (
            mock.patch.object(
                worker_allocator,
                "limit_by_max_active_instances",
                side_effect=lambda _, amount: amount,
            ),
            mock.patch.object(
                worker_allocator,
                "limit_by_pool_target_max_seconds_per_month",
                side_effect=lambda _, amount: amount,
            ),
            mock.patch.object(
                allocator,
                "limit_by_scope_pool_target_max_seconds_per_month",
                side_effect=limited_to,
            ),
        ):
            for a, e in zip(asked, expected):
                with mock.patch.object(
                    pool, "launch_workers"
                ) as launch_workers:
                    self.assertEqual(allocator.allocate_workers(pool, a), e)
                if e == 0:
                    launch_workers.assert_not_called()
                else:
                    launch_workers.assert_called_with(e)

    def test_pending_task_count(self) -> None:
        other_scope = self.playground.get_or_create_scope(name="other")
        other_workspace = self.playground.create_workspace(
            name="other", scope=other_scope, public=True
        )
        other_workspace_artifact = self.playground.create_source_artifact(
            workspace=other_workspace
        )
        other_environment = self.playground.create_debian_environment(
            workspace=other_workspace
        ).artifact
        assert other_environment is not None
        self.playground.advance_work_request(
            self.playground.create_sbuild_work_request(
                source=other_workspace_artifact,
                architecture="amd64",
                environment=other_environment,
                workspace=other_workspace,
            ),
            mark_pending=True,
        )

        artifact = self.playground.create_source_artifact()
        environment = self.playground.create_debian_environment().artifact
        assert environment is not None
        for architecture in ("amd64", "amd64", "ppc64el"):
            self.playground.advance_work_request(
                self.playground.create_sbuild_work_request(
                    source=artifact,
                    architecture=architecture,
                    environment=environment,
                ),
                mark_pending=True,
            )

        worker_allocator = WorkerAllocator()
        allocator = ScopeWorkerAllocator(worker_allocator, self.scenario.scope)
        self.assertEqual(
            allocator.pending_task_count(),
            {
                # Only amd64 tasks for the selected scope are counted
                "amd64": 2,
                "ppc64el": 1,
            },
        )

    def test_workers_required(self) -> None:
        worker_allocator = WorkerAllocator()
        allocator = ScopeWorkerAllocator(worker_allocator, self.scenario.scope)
        # We currently assume a task takes 10 minutes to run, and all tasks
        # need to run within an hour
        for queue_size, workers_needed in (
            (0, 0),  # Nothing to do
            (1, 1),  # One task can run within an hour, one worker is needed
            (7, 1),  # For up to 6 tasks, one worker is ok
            (8, 2),  # For the 7th task, we need another worker
            (12, 2),  # Two workers can do 12 tasks in an hour
        ):
            with (
                self.subTest(queue_size=queue_size),
                mock.patch.object(
                    allocator,
                    "pending_task_count",
                    return_value={"amd64": queue_size},
                ),
            ):
                # Calculations are based on the effort required to run all
                # but the last pending task.
                self.assertEqual(
                    allocator.workers_required(), {"amd64": workers_needed}
                )

    def test_workers_required_multi_archs(self) -> None:
        worker_allocator = WorkerAllocator()
        allocator = ScopeWorkerAllocator(worker_allocator, self.scenario.scope)
        with mock.patch.object(
            allocator,
            "pending_task_count",
            return_value={
                "amd64": 10,
                "arm64": 1,
                "i386": 0,
            },
        ):
            self.assertEqual(
                allocator.workers_required(),
                {
                    "amd64": 2,
                    "arm64": 1,
                    "i386": 0,
                },
            )

    def test_workers_available(self) -> None:
        provider_account = self.playground.create_cloud_provider_account_asset()

        instance_wide_pool = self.playground.create_worker_pool(
            "iw",
            instance_wide=True,
            provider_account=provider_account,
            architectures=["amd64", "i386"],
        )

        scope_pool = self.playground.create_worker_pool(
            "scope",
            instance_wide=False,
            provider_account=provider_account,
            architectures=["amd64", "i386"],
        )
        ScopeWorkerPool.objects.create(
            worker_pool=scope_pool,
            scope=self.scenario.scope,
        )

        other_pool = self.playground.create_worker_pool(
            "other",
            instance_wide=False,
            provider_account=provider_account,
            architectures=["amd64", "i386"],
        )

        for qty, connected, tombstone, pool, dyn_archs in (
            # 3 static workers, of which 1 is connected
            (1, True, False, None, ["amd64"]),
            (2, False, False, None, ["amd64"]),
            # 2 workers in instance_wide pool
            (2, False, False, instance_wide_pool, None),
            # 3 tombstones in instance_wide pool
            (3, False, True, instance_wide_pool, None),
            # 4 workers in scope_pool
            (4, False, False, scope_pool, None),
            # 8 workers in other_pool
            (8, False, False, other_pool, None),
        ):
            for _ in range(qty):
                worker = self.playground.create_worker(
                    worker_pool=pool,
                    extra_dynamic_metadata=(
                        {"system:architectures": dyn_archs}
                        if dyn_archs
                        else None
                    ),
                )
                if connected:
                    worker.mark_connected()
                if tombstone:
                    worker.instance_created_at = None
                    worker.save()

        worker_allocator = WorkerAllocator()
        allocator = ScopeWorkerAllocator(worker_allocator, self.scenario.scope)
        # 1 static (amd64-only), 2 instance_wide, 4 scope_pool
        self.assertEqual(
            allocator.workers_available(),
            {
                "amd64": 7,
                "i386": 6,
            },
        )

    def test_allocation_pools(self) -> None:
        provider_account = self.playground.create_cloud_provider_account_asset()

        instance_wide_pool1 = self.playground.create_worker_pool(
            "iw1", instance_wide=True, provider_account=provider_account
        )
        self.playground.create_worker_pool(
            "iw2",
            instance_wide=True,
            provider_account=provider_account,
            enabled=False,
        )

        scope_pool1 = self.playground.create_worker_pool(
            "scope1", instance_wide=False, provider_account=provider_account
        )
        ScopeWorkerPool.objects.create(
            worker_pool=scope_pool1,
            scope=self.scenario.scope,
            priority=1,
        )

        scope_pool2 = self.playground.create_worker_pool(
            "scope2", instance_wide=False, provider_account=provider_account
        )
        ScopeWorkerPool.objects.create(
            worker_pool=scope_pool2,
            scope=self.scenario.scope,
            priority=2,
        )

        scope_pool3 = self.playground.create_worker_pool(
            "scope3",
            instance_wide=False,
            provider_account=provider_account,
            enabled=False,
        )
        ScopeWorkerPool.objects.create(
            worker_pool=scope_pool3,
            scope=self.scenario.scope,
            priority=1,
        )

        self.playground.create_worker_pool(
            "other", instance_wide=False, provider_account=provider_account
        )

        worker_allocator = WorkerAllocator()
        allocator = ScopeWorkerAllocator(worker_allocator, self.scenario.scope)
        self.assertEqual(
            allocator.allocation_pools,
            [scope_pool2, scope_pool1, instance_wide_pool1],
        )

    def test_process_scope(self) -> None:
        """Test how process_scope allocates needed workers."""
        worker_allocator = WorkerAllocator()
        for required, available, npools, allocs_asked, allocs_given in (
            # 5 workers to allocate, one pool that can satisfy them all
            (10, 5, 1, [5], [5]),
            # 5 workers to allocate, first pool can satisfy them all
            (10, 5, 2, [5], [5]),
            # 5 workers to allocate, first pool can only allocate 2
            (10, 5, 2, [5, 2], [3, 2]),
            # No new workers to allocate
            (1, 2, 1, [], []),
        ):
            allocator = ScopeWorkerAllocator(
                worker_allocator, self.scenario.scope
            )
            allocator.allocation_pools = [
                WorkerPool(architectures=["amd64"]) for _ in range(npools)
            ]

            def fake_allocate(
                worker_pool: WorkerPool,  # noqa: ARG001, U100
                amount: int,  # noqa: ARG001, U100
            ) -> int:
                nonlocal available
                available += allocs_given[0]
                return allocs_given.pop(0)

            def fake_available() -> dict[str, int]:
                return {"amd64": available}

            with (
                mock.patch.object(
                    allocator, "allocate_workers", side_effect=fake_allocate
                ) as allocate_workers,
                mock.patch.object(
                    allocator,
                    "workers_required",
                    return_value={"amd64": required},
                ),
                mock.patch.object(
                    allocator, "workers_available", side_effect=fake_available
                ),
            ):
                allocator.process_scope()

            if not allocs_asked:
                allocate_workers.assert_not_called()
            else:
                for call, pool, asked in zip(
                    allocate_workers.call_args_list,
                    allocator.allocation_pools,
                    allocs_asked,
                ):
                    self.assertEqual(call.args, (pool, asked))

            # All expected allocations were made
            self.assertEqual(allocs_given, [])


class ScrubTests(TestCase):
    """Test :py:func:`scrub`."""

    def test_calls_scrub_workers(self) -> None:
        with mock.patch(
            "debusine.server.provisioning.scrub_workers",
        ) as scrub_workers:
            scrub()
        scrub_workers.assert_called()

    def test_uses_local_context(self) -> None:
        """
        Scrubbing uses a local context.

        Terminating workers may need to retry running work requests, which
        computes dynamic data, which may add entries to the lookup cache.
        Those entries do not persist between provisioning runs.
        """
        artifact, _ = self.playground.create_artifact()

        def fake_scrub_workers() -> None:
            lookup_single(
                f"{artifact.id}@artifacts",
                artifact.workspace,
                pc=PermissionContext.for_ui_user(
                    self.playground.get_default_user()
                ),
            )

        with (
            mock.patch(
                "debusine.server.provisioning.scrub_workers",
                side_effect=fake_scrub_workers,
            ),
            mock.patch("debusine.server.provisioning.create_new_workers"),
        ):
            scrub()

        self.assertEqual(context.lookup_cache, LookupCache())


class ScrubWorkersTests(TestCase):
    """Test :py:func:`scrub`."""

    def test_calls_pool_terminate_missing(self) -> None:
        self.playground.create_worker_pool()
        with mock.patch(
            "debusine.server.worker_pools.dummy.DummyWorkerPool"
            ".terminate_missing"
        ) as terminate_missing:
            scrub_workers()
        terminate_missing.assert_called()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Ontology definition of SystemBootstrap."""

import abc
import re
import shlex
import subprocess
import tempfile
from pathlib import Path
from typing import Any, cast, override
from urllib.parse import urljoin

import requests
from debian.deb822 import Deb822, Release

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.client.client_utils import get_url_contents_sha256sum
from debusine.client.exceptions import ContentValidationError
from debusine.tasks import RunCommandTask
from debusine.tasks.inputs import BuildArchitectureInput, Stage
from debusine.tasks.models import (
    BaseDynamicTaskData,
    SystemBootstrapData,
    SystemBootstrapRepository,
    SystemBootstrapRepositoryCheckSignatureWith,
    SystemBootstrapRepositoryKeyring,
)
from debusine.utils import get_user_agent
from debusine.utils.urls import DEFAULT_REQUEST_TIMEOUT


class SystemBootstrap[SBD: SystemBootstrapData](abc.ABC, RunCommandTask[SBD]):
    """Implement ontology SystemBootstrap."""

    TASK_VERSION = 1

    _OS_RELEASE_FILE = "os-release"

    build_architecture = BuildArchitectureInput(
        field="bootstrap_options.architecture", stage=Stage.PENDING
    )

    @override
    def post_init(self) -> None:
        super().post_init()

        # keyrings that will be uploaded into the chroot
        self._upload_keyrings: list[Path] = []

        # all the keyrings that have been downloaded
        self._keyrings: list[Path] = []

        self._host_sources_file: Path | None = None
        self._chroot_sources_file: Path | None = None

        # customization_script file
        self._customization_script: Path | None = None

    @staticmethod
    def _download_key(
        repository: SystemBootstrapRepository,
        *,
        keyring_directory: Path,
    ) -> str:
        # Using cast because repository's validator ensures keyring is set
        # when using external repositories
        repo_keyring = cast(
            SystemBootstrapRepositoryKeyring, repository.keyring
        )
        keyring, actual_sha256sum = get_url_contents_sha256sum(
            str(repo_keyring.url), 100 * 1024 * 1024, allow_file=True
        )

        if expected := repo_keyring.sha256sum:
            if actual_sha256sum != expected:
                raise ContentValidationError(
                    f"sha256 mismatch for keyring repository "
                    f"{repository.mirror}. "
                    f"Actual: {actual_sha256sum} expected: {expected}"
                )

        # Disable auto-deletion: the file will be deleted together
        # with keyring_directory when the RunCommandTask finishes
        file = tempfile.NamedTemporaryFile(  # noqa: SIM115
            dir=keyring_directory,
            prefix="keyring-repo-",
            # Detect ASCII armoring:
            # https://www.rfc-editor.org/rfc/rfc4880#section-6.2
            suffix=".asc" if keyring.startswith(b"-----BEGIN") else ".gpg",
            delete=False,
        )
        Path(file.name).write_bytes(keyring)
        return file.name

    @classmethod
    def _deb822_source(
        cls,
        repository: SystemBootstrapRepository,
        *,
        keyring_directory: Path,
        use_signed_by: bool,
        http_timeout: int | tuple[int, int] = DEFAULT_REQUEST_TIMEOUT,
    ) -> Deb822:
        """
        Create a deb822 from repository and return it.

        :raise: ContentValidationError if the repository["keyring"]["sha256sum"]
          does not match the one from the downloaded keyring
        :param keyring_directory: directory to save the gpg keys
        :param use_signed_by: add Signed-By in the repository
        :param http_timeout: Timeout for HTTP requests (connect, read).
        :return: repository
        """
        deb822 = Deb822()

        deb822["Types"] = " ".join(repository.types)
        deb822["URIs"] = repository.mirror
        deb822["Suites"] = repository.suite

        if (components := repository.components) is None:
            components = cls._list_components_for_suite(
                repository.mirror, repository.suite, timeout=http_timeout
            )

        deb822["Components"] = " ".join(components)

        if repository.check_signature_with == "no-check":
            deb822["Trusted"] = "yes"

        if (
            repository.check_signature_with
            == SystemBootstrapRepositoryCheckSignatureWith.EXTERNAL
        ):
            key_file = SystemBootstrap._download_key(
                repository, keyring_directory=keyring_directory
            )
            if use_signed_by:
                deb822["Signed-By"] = key_file

        return deb822

    @classmethod
    def _write_deb822s(cls, deb822s: list[Deb822], destination: Path) -> None:
        with destination.open("wb") as f:
            for deb822 in deb822s:
                deb822.dump(f)
                f.write(b"\n")

    def _generate_deb822_sources(
        self,
        repositories: list[SystemBootstrapRepository],
        *,
        keyrings_dir: Path,
        use_signed_by: bool,
    ) -> list[Deb822]:
        """
        Return list of Deb822 repositories.

        :param keyrings_dir: write gpg keys into it.
        :use_signed_by: if True, add Signed-By to the repository with the path
          to the file.
        """
        return [
            self._deb822_source(
                repository,
                keyring_directory=keyrings_dir,
                use_signed_by=use_signed_by,
                http_timeout=self.http_timeout or DEFAULT_REQUEST_TIMEOUT,
            )
            for repository in repositories
        ]

    @classmethod
    def _list_components_for_suite(
        cls,
        mirror_url: str,
        suite: str,
        *,
        timeout: int | tuple[int, int] = DEFAULT_REQUEST_TIMEOUT,
    ) -> list[str]:
        """
        Return components listed in the Release file.

        :raises ValueError: if components cannot be found or invalid names.
        """
        release_url = urljoin(mirror_url + "/", f"dists/{suite}/Release")
        try:
            headers = {"User-Agent": get_user_agent()}
            response = requests.get(
                release_url, headers=headers, timeout=timeout
            )
            components = Release(response.iter_lines())["components"].split()
        except KeyError:
            raise ValueError(f"Cannot find components in {release_url}")

        assert isinstance(components, list)
        for component in components:
            if re.search("^[A-Za-z][-/A-Za-z]*$", component) is None:
                raise ValueError(
                    f'Invalid component name from {release_url}: '
                    f'"{component}" must start with [A-Za-z] and have only '
                    f'[-/A-Za-z] characters'
                )

        return components

    @override
    def get_subject(self) -> str | None:
        return self.data.bootstrap_repositories[0].suite

    @override
    def get_configuration_context(self) -> str | None:
        return self.data.bootstrap_options.architecture

    @override
    def compute_scheduler_tags(
        self,
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        provided, required = super().compute_scheduler_tags()

        required.add(
            wtags.ProvenanceRequired.SYSTEM,
            {wtags.BUILD_ARCH_PREFIX + self.build_architecture},
        )

        return provided, required

    @override
    def can_run_on(self, worker_metadata: dict[str, Any]) -> bool:
        """Check if the specified worker can run the task."""
        if not super().can_run_on(worker_metadata):
            return False

        return self.build_architecture in worker_metadata.get(
            "system:architectures", []
        )

    @override
    def compute_dynamic_data(self) -> BaseDynamicTaskData:
        """
        Compute dynamic data.

        :subject: suite name in the first ``bootstrap_repositories``
        :parameter_summary: subject (architecture-variant)
        :runtime_context: ``$architecture:$variant:$extra_packages`` of
          ``bootstrap_options``
        :configuration_context: ``architecture`` of ``bootstrap_options``
        """
        suite_name = self.data.bootstrap_repositories[0].suite

        architecture = self.data.bootstrap_options.architecture
        variant = self.data.bootstrap_options.variant
        extra_packages = ",".join(
            sorted(self.data.bootstrap_options.extra_packages)
        )
        if variant is None:
            parameter_summary = f"{suite_name} ({architecture})"
        else:
            parameter_summary = f"{suite_name} ({architecture}-{variant})"

        return BaseDynamicTaskData(
            subject=suite_name,
            parameter_summary=parameter_summary,
            runtime_context=f"{architecture}:{variant}:{extra_packages}",
            configuration_context=architecture,
        )

    @override
    def configure_for_execution(self, download_dir: Path) -> bool:
        """Create file.sources and add it into the log."""
        # Create file.sources (keyrings are downloaded, if needed)
        keyrings_dir = download_dir / "keyrings"

        # mmdebstrap uses the subuid for using the keyring files. The directory
        # where the keyring files are saved must be accessible for the
        # mmdebstrap subuid user (and not only by the debusine-worker user).
        download_dir.chmod(0o755)
        keyrings_dir.mkdir(mode=0o755)

        host_sources = self._generate_deb822_sources(
            self.data.bootstrap_repositories,
            keyrings_dir=keyrings_dir,
            use_signed_by=self.data.bootstrap_options.use_signed_by,
        )

        # Make the files readable by any user (same reason as
        # above for the directory) and add them in self._keyrings
        # so cmdline use them
        for keyring in keyrings_dir.iterdir():
            keyring.chmod(0o644)
            self._keyrings.append(keyring)

        chroot_sources = [host_source.copy() for host_source in host_sources]

        # Change the path of the keyrings from the host paths to the
        # chroot path
        for chroot_source, task_repository in zip(
            chroot_sources, self.data.bootstrap_repositories
        ):
            if (signed_by := chroot_source.get("Signed-By")) is None:
                # Nothing needs to be done
                continue

            # If Signed-By has been set by _generate_deb822_sources, it means
            # that the repository specified a keyring
            assert task_repository.keyring is not None

            if task_repository.keyring.install:
                signed_by_path = Path(signed_by)
                filename = signed_by_path.name
                chroot_source["Signed-By"] = (
                    "/etc/apt/keyrings-debusine/" + filename
                )
                self._upload_keyrings.append(signed_by_path)
            else:
                # The key is not installed in the chroot, no "Signed-By"
                del chroot_source["Signed-By"]

        self._host_sources_file = download_dir / "host.sources"
        self._chroot_sources_file = download_dir / "chroot.sources"

        self._write_deb822s(host_sources, self._host_sources_file)
        self._write_deb822s(chroot_sources, self._chroot_sources_file)

        # Add the host and chroot source files for debugging purposes
        self.append_to_log_file(
            self._host_sources_file.name,
            self._host_sources_file.read_text().splitlines(),
        )

        self.append_to_log_file(
            self._chroot_sources_file.name,
            self._chroot_sources_file.read_text().splitlines(),
        )

        if script := self.data.customization_script:
            self._customization_script = download_dir / "customization_script"
            self._customization_script.write_text(script)

        return True

    @staticmethod
    def _get_source_components(apt_sources: Path) -> list[str]:
        """Return the components for the first APT source in apt_sources."""
        for source in Deb822.iter_paragraphs(apt_sources.read_text()):
            components = cast(str, source.get("Components", ""))
            return components.split()
        return []

    @staticmethod
    def _get_value_os_release(os_release: Path, key: str) -> str:
        """Parse file with the format /etc/os-release, return value for key."""
        # In https://www.freedesktop.org/software/systemd/man/latest/os-release.html  # noqa: E501
        # it specifies that the file is a "newline-separated list of
        # environment-like shell-compatible variable assignments".
        # Thus, shlex.split() is used it removes the quotes if they are in
        # any value

        for key_value in shlex.split(os_release.read_text()):
            k, v = key_value.split("=", 1)

            if key == k:
                return v

        raise KeyError(key)

    @staticmethod
    def _get_pkglist(var_lib_dpkg: Path) -> dict[str, str]:
        """
        Execute dpkg-query --admindir=var_lib_dpkg/dpkg --show.

        :return: dictionary with package names and versions.
        """
        cmd = [
            "dpkg-query",
            f"--admindir={var_lib_dpkg}/dpkg",
            "--show",
        ]

        process = subprocess.run(
            cmd, check=True, text=True, capture_output=True
        )

        result = {}
        for line in process.stdout.splitlines():
            name, version = line.split("\t")
            result[name] = version

        return result



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for QA work request plugin."""

from typing import ClassVar

from debusine.db.models import WorkRequest
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.views.qa import QATab, QAViewWorkRequestPlugin
from debusine.web.views.work_request import Tab


class QAViewWorkRequestPluginTests(TestCase):
    """Tests for :py:class:`QAViewWorkRequestPlugin`."""

    scenario = scenarios.UIRegressionTesting(simulate=False)

    work_request: ClassVar[WorkRequest]

    def plugin(self) -> QAViewWorkRequestPlugin:
        return QAViewWorkRequestPlugin(self.scenario.workflow_qa.get_task())

    def get_tab(self) -> QATab:
        """Return the tab for a work request or a plugin."""
        plugin = self.plugin()
        tabs: list[Tab] = []
        plugin.add_tabs(tabs)
        self.assertEqual(len(tabs), 1)
        tab = tabs[0]
        assert isinstance(tab, QATab)
        self.assertEqual(tab.name, "qa")
        self.assertEqual(tab.label, "QA")
        return tab

    def test_is_enabled(self) -> None:
        plugin = self.plugin()
        self.assertTrue(plugin.is_enabled())

    def test_is_enabled_requires_output_data(self) -> None:
        self.scenario.workflow_qa.output_data = None
        self.scenario.workflow_qa.save()
        plugin = self.plugin()
        self.assertFalse(plugin.is_enabled())

    def test_is_enabled_requires_regression_analysis(self) -> None:
        od = self.scenario.workflow_qa.output_data
        assert od is not None
        od.regression_analysis = None
        self.scenario.workflow_qa.output_data = od
        self.scenario.workflow_qa.save()
        plugin = self.plugin()
        self.assertFalse(plugin.is_enabled())

    def test_is_enabled_requires_regression_analysis_summary(self) -> None:
        od = self.scenario.workflow_qa.output_data
        assert od is not None
        ra = od.regression_analysis
        assert ra is not None
        del ra[""]
        self.scenario.workflow_qa.output_data = od
        self.scenario.workflow_qa.save()
        plugin = self.plugin()
        self.assertFalse(plugin.is_enabled())

    def test_get_context_data_only_summary(self) -> None:
        od = self.scenario.workflow_qa.output_data
        assert od is not None
        ra = od.regression_analysis
        assert ra is not None
        ctx = self.get_tab().get_context_data()
        self.assertEqual(ctx["summary"], ra[""])
        self.assertEqual(ctx["orig"], self.scenario.reference_source)
        self.assertEqual(ctx["new"], self.scenario.new_source)
        self.assertEqual(ctx["rows"], [])
        self.assertEqual(ctx["warnings"], [])

    def test_get_context_data_rows(self) -> None:
        self.scenario.simulate_sbuild()
        self.scenario.simulate_autopkgtest()
        od = self.scenario.workflow_qa.output_data
        assert od is not None
        ra = od.regression_analysis
        assert ra is not None
        ctx = self.get_tab().get_context_data()
        self.assertEqual(ctx["summary"], ra[""])
        self.assertEqual(ctx["orig"], self.scenario.reference_source)
        self.assertEqual(ctx["new"], self.scenario.new_source)
        self.assertEqual(
            ctx["rows"],
            [
                {
                    "name": "autopkgtest:libfoo:amd64",
                    "orig": None,
                    "new": self.scenario.autopkgtest_libfoo_new_amd64,
                    "analysis": self.scenario.ra_autopkgtest_amd64,
                },
                {
                    "name": "autopkgtest:libfoo:i386",
                    "orig": self.scenario.autopkgtest_libfoo_ref_i386,
                    "new": self.scenario.autopkgtest_libfoo_new_i386,
                    "analysis": self.scenario.ra_autopkgtest_i386,
                },
            ],
        )
        self.assertEqual(ctx["warnings"], [])

    def test_get_context_data_rows_hide_boring_revdep(self) -> None:
        self.scenario.simulate_sbuild()
        self.scenario.simulate_rev_dep_autopkgtest()
        od = self.scenario.workflow_qa.output_data
        assert od is not None
        ra = od.regression_analysis
        assert ra is not None
        ctx = self.get_tab().get_context_data()
        self.assertEqual(ctx["summary"], ra[""])
        self.assertEqual(ctx["orig"], self.scenario.reference_source)
        self.assertEqual(ctx["new"], self.scenario.new_source)
        self.assertEqual(
            ctx["rows"],
            [
                # autopkgtest:foo:amd64 is hidden because it is STABLE
                {
                    "name": "autopkgtest:foo:i386",
                    "orig": self.scenario.autopkgtest_foo_ref_i386,
                    "new": self.scenario.autopkgtest_foo_new_i386,
                    "analysis": self.scenario.ra_autopkgtest_foo_i386,
                },
            ],
        )
        self.assertEqual(ctx["warnings"], [])

    def test_get_context_data_missing_artifacts(self) -> None:
        self.scenario.simulate_sbuild()
        self.scenario.simulate_autopkgtest()
        od = self.scenario.workflow_qa.output_data
        assert od is not None
        ra = od.regression_analysis
        assert ra is not None
        ra[""].original_artifact_id = None
        ra[""].new_artifact_id = None
        self.scenario.workflow_qa.output_data = od
        self.scenario.workflow_qa.save()
        ctx = self.get_tab().get_context_data()
        self.assertEqual(ctx["summary"], ra[""])
        self.assertIsNone(ctx["orig"])
        self.assertIsNone(ctx["new"])
        self.assertEqual(ctx["warnings"], [])

    def test_render_no_rows(self) -> None:
        response = self.client.get(self.scenario.workflow_qa.get_absolute_url())
        tree = self.assertResponseHTML(response, dump_on_error=True)
        tab = self.assertHasElement(tree, "//div[@id='nav-qa']")
        self.assertTextContentEqual(tab, "No regression analysis results")

    def test_render(self) -> None:
        self.scenario.simulate_sbuild()
        self.scenario.simulate_autopkgtest()
        od = self.scenario.workflow_qa.output_data
        assert od is not None
        ra = od.regression_analysis
        assert ra is not None
        response = self.client.get(self.scenario.workflow_qa.get_absolute_url())
        tree = self.assertResponseHTML(response, dump_on_error=True)
        tab = self.assertHasElement(tree, "//div[@id='nav-qa']")
        thead = self.assertHasElement(tab, "table/thead")
        self.assertTextContentEqual(
            thead.tr[0].th[1],
            f"Original result for {ra[""].original_source_version}",
        )
        self.assertTextContentEqual(
            thead.tr[0].th[2], f"New result for {ra[""].new_source_version}"
        )
        tbody = self.assertHasElement(tab, "table/tbody")
        self.assertTextContentEqual(
            tbody.tr[0].td[0], "autopkgtest:libfoo:amd64"
        )
        self.assertFalse(tbody.tr[0].td[1].xpath("a"))
        self.assertTextContentEqual(tbody.tr[0].td[1], "-")
        a = self.assertHasElement(tbody.tr[0].td[2], "a")
        self.assertEqual(
            a.get("href"),
            self.scenario.autopkgtest_libfoo_new_amd64.get_absolute_url(),
        )
        self.assertTextContentEqual(tbody.tr[0].td[2], "Success")
        self.assertTextContentEqual(tbody.tr[0].td[3], "? no-result")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Executors for Debusine Tasks.

Provides containment for tasks using a variety of backends.
"""

from debusine.tasks.executors.base import (
    ExecutorImageCategory,
    ExecutorInterface,
    ExecutorStatistics,
    InstanceInterface,
    analyze_worker_all_executors,
    executor_backends,
    executor_class,
    executor_provided_tags,
)
from debusine.tasks.executors.incus import (
    IncusInstance,
    IncusLXCExecutor,
    IncusVMExecutor,
)
from debusine.tasks.executors.qemu import QemuExecutor
from debusine.tasks.executors.unshare import UnshareExecutor, UnshareInstance

__all__ = [
    "ExecutorImageCategory",
    "ExecutorInterface",
    "ExecutorStatistics",
    "IncusInstance",
    "IncusLXCExecutor",
    "IncusVMExecutor",
    "InstanceInterface",
    "QemuExecutor",
    "UnshareExecutor",
    "UnshareInstance",
    "analyze_worker_all_executors",
    "executor_backends",
    "executor_class",
    "executor_provided_tags",
]



# Generated by Django 4.2.23 on 2026-02-19 19:13

from django.db import migrations, models

import debusine.assets.models
import debusine.db.constraints


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0042_add_scheduler_tags_for_tasks'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='asset',
            name='db_asset_workspace_not_null',
        ),
        migrations.AlterField(
            model_name='asset',
            name='category',
            field=models.CharField(
                choices=[
                    (
                        debusine.assets.models.AssetCategory[
                            'CLOUD_PROVIDER_ACCOUNT'
                        ],
                        debusine.assets.models.AssetCategory[
                            'CLOUD_PROVIDER_ACCOUNT'
                        ],
                    ),
                    (
                        debusine.assets.models.AssetCategory['SIGNING_KEY'],
                        debusine.assets.models.AssetCategory['SIGNING_KEY'],
                    ),
                    (
                        debusine.assets.models.AssetCategory[
                            'APT_AUTHENTICATION'
                        ],
                        debusine.assets.models.AssetCategory[
                            'APT_AUTHENTICATION'
                        ],
                    ),
                ],
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name='assetusagerole',
            name='role',
            field=models.CharField(
                choices=[
                    ('signer', 'Signer'),
                    ('repository_signer', 'Repository signer'),
                    ('apt_authenticator', 'Can use APT authentication'),
                ],
                max_length=32,
            ),
        ),
        migrations.AddConstraint(
            model_name='asset',
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(
                        (
                            'category',
                            debusine.assets.models.AssetCategory[
                                'APT_AUTHENTICATION'
                            ],
                        ),
                        _negated=True,
                    ),
                    ('data__name__regex', '^[A-Za-z][A-Za-z0-9+._-]*$'),
                    _connector='OR',
                ),
                name='db_asset_apt_auth_name',
            ),
        ),
        migrations.AddConstraint(
            model_name='asset',
            constraint=debusine.db.constraints.JsonDataUniqueConstraint(
                condition=models.Q(
                    (
                        'category',
                        debusine.assets.models.AssetCategory[
                            'APT_AUTHENTICATION'
                        ],
                    )
                ),
                fields=('workspace', "data->>'name'"),
                name='db_asset_unique_apt_auth_workspace_name',
                nulls_distinct=False,
            ),
        ),
        migrations.AddConstraint(
            model_name='asset',
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(
                        (
                            'category__in',
                            {
                                debusine.assets.models.AssetCategory[
                                    'APT_AUTHENTICATION'
                                ],
                                debusine.assets.models.AssetCategory[
                                    'SIGNING_KEY'
                                ],
                            },
                        ),
                        _negated=True,
                    ),
                    ('workspace__isnull', False),
                    _connector='OR',
                ),
                name='db_asset_workspace_not_null',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Debusine signing settings.

The settings are created dynamically by first importing defaults
values from :py:mod:`debusine.signing.settings.defaults` and then
values from :py:mod:`debusine.signing.settings.local` (or from
:py:mod:`debusine.signing.settings.selected` if the latter
has not been created by the administrator).
"""

from debusine.django.django_utils import in_test_suite
from debusine.signing.settings.defaults import *  # noqa: F403

if in_test_suite():
    from debusine.signing.settings.test import *  # noqa: F403
else:
    try:
        from debusine.signing.settings.local import *  # noqa: F403
    except ModuleNotFoundError as e:
        if e.name != "debusine.signing.settings.local":
            raise
        from debusine.signing.settings.selected import *  # noqa: F403

compute_default_settings(globals())  # noqa: F405



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Manager for debusine:workflow-internal collections."""

import datetime as dt
from typing import Any, override

from django.db import IntegrityError
from django.utils import timezone

from debusine.artifacts.models import BareDataCategory, CollectionCategory
from debusine.db.models import Artifact, CollectionItem, User, WorkRequest
from debusine.server.collections.base import (
    CollectionManagerInterface,
    ItemAdditionError,
)


class WorkflowInternalManager(CollectionManagerInterface):
    """
    Manage collection of category debusine:workflow-internal.

    This collection stores runtime data of a workflow.  Bare items can be
    used to store arbitrary JSON data, while artifact items can help to
    share artifacts between all the tasks (and help retain them for
    long-running workflows).
    """

    COLLECTION_CATEGORY = CollectionCategory.WORKFLOW_INTERNAL
    VALID_BARE_DATA_CATEGORIES = None
    VALID_ARTIFACT_CATEGORIES = None

    @override
    def do_add_bare_data(
        self,
        category: BareDataCategory,
        *,
        user: User,
        workflow: WorkRequest | None = None,
        data: dict[str, Any] | None = None,  # noqa: U100
        name: str | None = None,
        replace: bool = False,
        created_at: dt.datetime | None = None,
        replaced_by: CollectionItem | None = None,
    ) -> CollectionItem:
        if name is None:
            raise ItemAdditionError(
                f"Adding to {CollectionCategory.WORKFLOW_INTERNAL} requires "
                f"an item name"
            )

        if data is not None and category != BareDataCategory.PROMISE:
            # Only PROMISE can have data keys starting with promise_
            for key in data:
                if key.startswith("promise_"):
                    raise ItemAdditionError(
                        f'Fields starting with "promise_" are not '
                        f'allowed for category "{category}"'
                    )

        if replace:
            self.remove_items_by_name(
                name=name,
                child_types=[CollectionItem.Types.BARE],
                user=user,
                workflow=workflow,
            )

        try:
            return CollectionItem.objects.create_from_bare_data(
                category,
                parent_collection=self.collection,
                name=name,
                data=data or {},
                created_at=created_at,
                created_by_user=user,
                created_by_workflow=workflow,
                replaced_by=replaced_by,
            )
        except IntegrityError as exc:
            raise ItemAdditionError(str(exc))

    @override
    def do_add_artifact(
        self,
        artifact: Artifact,
        *,
        user: User,
        workflow: WorkRequest | None = None,
        variables: dict[str, Any] | None = None,
        name: str | None = None,
        replace: bool = False,
        created_at: dt.datetime | None = None,
        replaced_by: CollectionItem | None = None,
    ) -> CollectionItem:
        """
        Add the artifact into the managed collection.

        If replace=True the artifact can replace an artifact or bare data.
        """
        if name is None:
            raise ItemAdditionError(
                f"Adding to {CollectionCategory.WORKFLOW_INTERNAL} requires "
                f"an item name"
            )

        if replace:
            self.remove_items_by_name(
                name=name,
                child_types=[
                    CollectionItem.Types.ARTIFACT,
                    CollectionItem.Types.BARE,
                ],
                user=user,
                workflow=workflow,
            )

        if variables is not None:
            for key in variables:
                # Forbid names starting with 'promise_'
                if key.startswith("promise_"):
                    raise ItemAdditionError(
                        'Keys starting with "promise_" '
                        'are not allowed in variables'
                    )
        try:
            return CollectionItem.objects.create_from_artifact(
                artifact,
                parent_collection=self.collection,
                name=name,
                data=variables or {},
                created_at=created_at,
                created_by_user=user,
                created_by_workflow=workflow,
                replaced_by=replaced_by,
            )
        except IntegrityError as exc:
            raise ItemAdditionError(str(exc))

    @override
    def do_remove_item(
        self,
        item: CollectionItem,
        *,
        user: User | None = None,
        workflow: WorkRequest | None = None,
    ) -> None:
        item.removed_by_user = user
        item.removed_by_workflow = workflow
        item.removed_at = timezone.now()
        item.save()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Celery integration for debusine workflow orchestrators."""

import logging

from celery import shared_task
from django.db import transaction
from django.utils import timezone
from prometheus_client import Histogram

from debusine.artifacts.models import TaskTypes
from debusine.db.context import context
from debusine.db.metrics import emit_metric
from debusine.db.models import WorkRequest
from debusine.server.workflows.base import orchestrate_workflow
from debusine.tasks.models import OutputData, OutputDataError

logger = logging.getLogger(__name__)

workflow_orchestrator_duration = Histogram(
    "debusine_workflow_orchestrator_duration",
    "Histogram of how long workflow orchestrators took to run.",
    labelnames=("task_name",),
    # Orchestrators can take some time.  Try to provide a reasonable range
    # of buckets.
    buckets=(
        1.0,
        5.0,
        15.0,
        60.0,  # 1 minute
        300.0,  # 5 minutes
        900.0,  # 15 minutes
        3600.0,  # 1 hour
        10800.0,  # 3 hours
        21600.0,  # 6 hours
        43200.0,  # 12 hours
        86400.0,  # 1 day
        float("inf"),
    ),
)


# mypy complains that celery.shared_task is untyped, which is true, but we
# can't fix that here.
@shared_task  # type: ignore[untyped-decorator]
@transaction.atomic
def run_workflow_task(work_request_id: int) -> WorkRequest.Results:
    """Run a workflow callback or workflow in Celery."""
    try:
        work_request = (
            WorkRequest.objects.filter(id=work_request_id)
            .lock_workflows_for_update()
            .get()
        )
    except WorkRequest.DoesNotExist:
        logger.error("Work request %d does not exist", work_request_id)
        raise

    # orchestrate_workflow will mark the workflow running, but we need to
    # make sure that that was already done before dispatching the Celery
    # task in order that it's visible outside this transaction.
    if work_request.status != WorkRequest.Statuses.RUNNING:
        message = (
            f"Work request is in status {work_request.status}, not running"
        )
        logger.error(
            "Error running work request %s/%s (%s): %s",
            work_request.task_type,
            work_request.task_name,
            work_request.id,
            message,
        )
        work_request.mark_completed(
            WorkRequest.Results.ERROR,
            output_data=OutputData(
                errors=[OutputDataError(message=message, code="wrong-status")]
            ),
        )
        return WorkRequest.Results.ERROR

    if (
        work_request.task_type == TaskTypes.WORKFLOW
        and not work_request.is_workflow_root
    ):
        # This task must only be called to populate root workflows.
        # Sub-workflows are populated by their parent workflows instead.
        message = "Must be populated by its parent workflow instead"
        logger.error(
            "Error running work request %s/%s (%s): %s",
            work_request.task_type,
            work_request.task_name,
            work_request.id,
            message,
        )
        work_request.mark_completed(
            WorkRequest.Results.ERROR,
            output_data=OutputData(
                errors=[
                    OutputDataError(message=message, code="non-root-workflow")
                ]
            ),
        )
        return WorkRequest.Results.ERROR

    context.reset()
    work_request.set_current()
    start_time = timezone.now()

    result = (
        WorkRequest.Results.SUCCESS
        if orchestrate_workflow(work_request)
        else WorkRequest.Results.ERROR
    )

    emit_metric(
        metric_type="histogram",
        name="debusine_workflow_orchestrator_duration",
        labels={"task_name": work_request.task_name},
        value=(timezone.now() - start_time).total_seconds(),
    )

    return result



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for ExternalDebianSuiteFileBackend."""

from pathlib import Path
from typing import ClassVar, override

import responses

from debusine.db.models import FileStore
from debusine.db.tests.utils import _calculate_hash_from_data
from debusine.server.file_backend.external import ExternalDebianSuiteFileBackend
from debusine.test.django import TestCase
from debusine.utils import NotSupportedError


class ExternalDebianSuiteFileBackendTests(TestCase):
    """Tests for ExternalDebianSuiteFileBackend."""

    db_store: ClassVar[FileStore]
    file_backend: ClassVar[ExternalDebianSuiteFileBackend]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        """Initialize file backend to be tested."""
        super().setUpTestData()

        cls.db_store = FileStore.objects.create(
            name="ExternalDebianSuite",
            backend=FileStore.BackendChoices.EXTERNAL_DEBIAN_SUITE,
            configuration={
                "archive_root_url": "http://ftp.debian.org/debian/",
                "suite": "bookworm",
                "components": ["main", "contrib"],
            },
        )

        cls.file_backend = ExternalDebianSuiteFileBackend(cls.db_store)

    def test_entry_str(self) -> None:
        """ExternalFileBackendEntry.__str__ returns the URL."""
        file = self.file_backend.add_remote_file(
            "pool/main/c/coreutils/coreutils_9.1-1_amd64.deb",
            _calculate_hash_from_data(b"test"),
            len(b"test"),
        )
        entry = self.file_backend.get_entry(file)
        expected_url = (
            "http://ftp.debian.org/debian/"
            "pool/main/c/coreutils/coreutils_9.1-1_amd64.deb"
        )
        self.assertEqual(str(entry), expected_url)

    @responses.activate
    def test_get_temporary_local_path(self) -> None:
        """`get_temporary_local_path` yields a temporary local path."""
        db_file = self.file_backend.add_remote_file(
            "pool/main/c/coreutils/coreutils_9.1-1_amd64.deb",
            bytes.fromhex(
                "61038f857e346e8500adf53a2a0a2085"
                "9f4d3a3b51570cc876b153a2d51a3091"
            ),
            2896560,
        )
        expected_url = (
            "http://ftp.debian.org/debian/"
            "pool/main/c/coreutils/coreutils_9.1-1_amd64.deb"
        )
        mocked_content = b"coreutils.deb_mocked_content"
        responses.add(responses.GET, expected_url, mocked_content)

        with self.file_backend.get_temporary_local_path(db_file) as temp_path:
            self.assertEqual(temp_path.read_bytes(), mocked_content)

    @responses.activate
    def test_remote_file(self) -> None:  # pragma: no cover
        """Create a remote file."""
        db_file = self.file_backend.add_remote_file(
            "pool/main/c/coreutils/coreutils_9.1-1_amd64.deb",
            bytes.fromhex(
                "61038f857e346e8500adf53a2a0a2085"
                "9f4d3a3b51570cc876b153a2d51a3091"
            ),
            2896560,
        )

        expected_url = (
            "http://ftp.debian.org/debian/"
            "pool/main/c/coreutils/coreutils_9.1-1_amd64.deb"
        )
        self.assertEqual(
            self.file_backend.get_url(db_file),
            expected_url,
        )

        self.assertIsNone(self.file_backend.get_local_path(db_file))

        mocked_content = b"coreutils.deb_mocked_content"
        responses.add(
            responses.GET,
            expected_url,
            mocked_content,
        )

        content = b''
        with self.file_backend.get_stream(db_file) as stream:
            content += stream.read()
        self.assertEqual(content, mocked_content)

        with self.assertRaises(NotSupportedError):
            self.file_backend.remove_file(db_file)

        with self.assertRaises(NotSupportedError):
            self.file_backend.add_file(Path('a/b/c.deb'), db_file)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for UpdateEnvironmentsWorkflow."""

from typing import Any, override

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    TaskTypes,
)
from debusine.client.models import RuntimeParameter
from debusine.db.models import WorkRequest, WorkflowTemplate, default_workspace
from debusine.server.collections.tests.utils import CollectionTestMixin
from debusine.server.workflows import UpdateEnvironmentsWorkflow
from debusine.server.workflows.base import orchestrate_workflow
from debusine.server.workflows.tests.helpers import WorkflowTestCase
from debusine.tasks.models import (
    ActionUpdateCollectionWithArtifacts,
    BackendType,
    BaseDynamicTaskData,
)


class UpdateEnvironmentsWorkflowTests(
    CollectionTestMixin, WorkflowTestCase[UpdateEnvironmentsWorkflow]
):
    """Unit tests for UpdateEnvironmentsWorkflow."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.targets: list[dict[str, Any]] = []

    def create_update_environments_template(
        self,
        name: str = "update_environments",
        static_parameters: dict[str, Any] | None = None,
    ) -> WorkflowTemplate:
        """Create an update_environments WorkflowTemplate."""
        return WorkflowTemplate.objects.create(
            name=name,
            workspace=default_workspace(),
            task_name="update_environments",
            static_parameters=static_parameters or {},
            runtime_parameters=RuntimeParameter.ANY,
        )

    def add_target(
        self,
        codename: str = "bookworm",
        codenames: list[str] | None = None,
        aliases: list[str] | None = None,
        architectures: list[str] | None = None,
        backends: list[str] | None = None,
        variants: list[str | None] | None = None,
        mmdebstrap_template: dict[str, Any] | None = None,
        simplesystemimagebuild_template: dict[str, Any] | None = None,
    ) -> None:
        # Compute default values for parameters
        if codenames is None:
            codenames = [codename]
        if architectures is None:
            architectures = ["amd64"]
        if aliases is None:
            aliases = [f"{codenames[0]}-alias"]
        if (
            mmdebstrap_template is None
            and simplesystemimagebuild_template is None
        ):
            # Ensure we build something, defaults to tarball with mmdebstrap
            mmdebstrap_template = {}
        if mmdebstrap_template is not None:
            mmdebstrap_template.setdefault(
                "bootstrap_repositories",
                [{"mirror": "https://deb.debian.org/debian"}],
            )
        if simplesystemimagebuild_template is not None:
            simplesystemimagebuild_template.setdefault(
                "bootstrap_repositories",
                [{"mirror": "https://deb.debian.org/debian"}],
            )
            simplesystemimagebuild_template.setdefault(
                "disk_image",
                {
                    "format": "qcow2",
                    "partitions": [{"size": 2, "filesystem": "ext4"}],
                },
            )

        # Generate the target entry
        target: dict[str, Any] = {
            "codenames": codenames,
            "codename_aliases": {codenames[0]: aliases},
            "architectures": architectures,
            "mmdebstrap_template": mmdebstrap_template,
            "simplesystemimagebuild_template": simplesystemimagebuild_template,
        }
        if backends is not None:
            target["backends"] = backends
        if variants is not None:
            target["variants"] = variants

        self.targets.append(target)

    @override
    def work_request_create(self, **kwargs: Any) -> WorkRequest:
        template = self.create_update_environments_template(
            static_parameters={"vendor": "debian", "targets": self.targets}
        )
        work_request = self.playground.create_workflow(template, task_data={})
        self.assertEqual(work_request.status, WorkRequest.Statuses.PENDING)
        self.assertTrue(orchestrate_workflow(work_request))
        return work_request

    def test_create_orchestrator(self) -> None:
        """Instantiate an UpdateEnvironmentsWorkflow."""
        self.add_target(mmdebstrap_template={})
        w = self.task
        self.assertEqual(w.data.vendor, "debian")
        self.assertEqual(len(w.data.targets), 1)
        self.assertEqual(w.data.targets[0].codenames, ["bookworm"])
        self.assertEqual(w.data.targets[0].architectures, ["amd64"])
        assert w.data.targets[0].mmdebstrap_template is not None
        self.assertIsNotNone(
            w.data.targets[0].mmdebstrap_template.get("bootstrap_repositories")
        )

    def test_create_work_requests(self) -> None:
        """Create specified work requests."""
        self.add_target(mmdebstrap_template={}, aliases=[])
        self.add_target(
            codename="trixie",
            aliases=["sid"],
            architectures=["amd64", "arm64"],
            variants=[None, "autopkgtest", "sbuild"],
            backends=[
                BackendType.UNSHARE,
                BackendType.INCUS_LXC,
            ],
            simplesystemimagebuild_template={
                "bootstrap_options": {"variant": "minbase"}
            },
        )
        wr = self.work_request

        children = list(wr.children.order_by("id"))
        self.assertEqual(len(children), 3)

        expected_bookworm_task_data = {
            "bootstrap_repositories": [
                {"mirror": "https://deb.debian.org/debian", "suite": "bookworm"}
            ],
        }
        expected_bookworm_reactions = [
            ActionUpdateCollectionWithArtifacts(
                artifact_filters={"category": ArtifactCategory.SYSTEM_TARBALL},
                collection="debian@debian:environments",
                variables={"codename": "bookworm"},
            )
        ]

        self.assertEqual(children[0].status, WorkRequest.Statuses.PENDING)
        self.assertEqual(children[0].task_type, TaskTypes.WORKER)
        self.assertEqual(children[0].task_name, "mmdebstrap")
        self.assertEqual(
            children[0].task_data,
            {
                "bootstrap_options": {"architecture": "amd64"},
                **expected_bookworm_task_data,
            },
        )
        self.assertQuerySetEqual(children[0].dependencies.all(), [])
        self.assertEqual(
            children[0].workflow_data_json,
            {
                "display_name": (
                    "Build tarball for bookworm / amd64 / no-backend "
                    "/ no-variant"
                ),
                "step": "mmdebstrap/bookworm/amd64/no-backend/no-variant",
            },
        )
        self.assert_work_request_event_reactions(
            children[0], on_success=expected_bookworm_reactions
        )

        expected_trixie_task_data = {
            "bootstrap_repositories": [
                {
                    "mirror": "https://deb.debian.org/debian",
                    "suite": "trixie",
                }
            ],
            "disk_image": {
                "format": "qcow2",
                "partitions": [{"size": 2, "filesystem": "ext4"}],
            },
        }
        expected_trixie_reactions = [
            ActionUpdateCollectionWithArtifacts(
                artifact_filters={"category": ArtifactCategory.SYSTEM_IMAGE},
                collection="debian@debian:environments",
                variables={
                    "codename": codename,
                    "backend": backend,
                    **({} if variant is None else {"variant": variant}),
                },
            )
            for codename, variant, backend in (
                ("trixie", None, BackendType.UNSHARE),
                ("trixie", None, BackendType.INCUS_LXC),
                ("trixie", "autopkgtest", BackendType.UNSHARE),
                ("trixie", "autopkgtest", BackendType.INCUS_LXC),
                ("trixie", "sbuild", BackendType.UNSHARE),
                ("trixie", "sbuild", BackendType.INCUS_LXC),
                ("sid", None, BackendType.UNSHARE),
                ("sid", None, BackendType.INCUS_LXC),
                ("sid", "autopkgtest", BackendType.UNSHARE),
                ("sid", "autopkgtest", BackendType.INCUS_LXC),
                ("sid", "sbuild", BackendType.UNSHARE),
                ("sid", "sbuild", BackendType.INCUS_LXC),
            )
        ]

        self.assertEqual(children[1].status, WorkRequest.Statuses.PENDING)
        self.assertEqual(children[1].task_type, TaskTypes.WORKER)
        self.assertEqual(children[1].task_name, "simplesystemimagebuild")
        self.assertEqual(
            children[1].task_data,
            {
                "bootstrap_options": {
                    "architecture": "amd64",
                    "variant": "minbase",
                },
                **expected_trixie_task_data,
            },
        )
        self.assertQuerySetEqual(children[1].dependencies.all(), [])
        self.assertEqual(
            children[1].workflow_data_json,
            {
                "display_name": (
                    "Build image for trixie / amd64 / unshare|"
                    "incus-lxc / no-variant|autopkgtest|sbuild"
                ),
                "step": (
                    "simplesystemimagebuild/trixie/amd64/"
                    "unshare|incus-lxc/no-variant|autopkgtest|sbuild"
                ),
            },
        )
        self.assert_work_request_event_reactions(
            children[1], on_success=expected_trixie_reactions
        )

        self.assertEqual(children[2].status, WorkRequest.Statuses.PENDING)
        self.assertEqual(children[2].task_type, TaskTypes.WORKER)
        self.assertEqual(children[2].task_name, "simplesystemimagebuild")
        self.assertEqual(
            children[2].task_data,
            {
                "bootstrap_options": {
                    "architecture": "arm64",
                    "variant": "minbase",
                },
                **expected_trixie_task_data,
            },
        )
        self.assertQuerySetEqual(children[2].dependencies.all(), [])
        self.assertEqual(
            children[2].workflow_data_json,
            {
                "display_name": (
                    "Build image for trixie / arm64 / "
                    "unshare|incus-lxc / no-variant|autopkgtest|sbuild"
                ),
                "step": (
                    "simplesystemimagebuild/trixie/arm64/"
                    "unshare|incus-lxc/no-variant|autopkgtest|sbuild"
                ),
            },
        )
        self.assert_work_request_event_reactions(
            children[2], on_success=expected_trixie_reactions
        )

        # populate() is idempotent.
        orchestrator = self.get_workflow(wr)
        orchestrator.populate()
        self.assertEqual(wr.children.count(), 3)

    def test_event_reaction(self) -> None:
        """The event reaction created by the workflow can be handled."""
        self.add_target()
        wr = self.work_request

        child = WorkRequest.objects.get(parent=wr)
        child.status = WorkRequest.Statuses.PENDING
        collection = self.playground.create_collection(
            name="debian", category=CollectionCategory.ENVIRONMENTS
        )
        tarball, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL,
            data={"architecture": "amd64"},
            work_request=child,
        )

        self.playground.advance_work_request(
            child, result=WorkRequest.Results.SUCCESS
        )

        item = collection.manager.lookup(
            "match:format=tarball:codename=bookworm:architecture=amd64"
        )
        assert item is not None
        self.assertEqual(item.artifact, tarball)

    def test_compute_scheduler_tags(self) -> None:
        self.add_target()
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data(self) -> None:
        self.add_target(architectures=["amd64", "i386"])
        self.add_target(architectures=["arm64"])
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            BaseDynamicTaskData(
                subject="debian",
                parameter_summary="debian (3 environments)",
            ),
        )

    def test_combination_logic_single_target(self) -> None:
        self.add_target(
            codenames=["bullseye", "bookworm", "trixie", "sid"],
            architectures=["amd64", "arm64", "armhf"],
            mmdebstrap_template={},
            simplesystemimagebuild_template={},
        )
        # 4 codenames, 3 architectures, 2 tasks (tarball+image)
        self.assertEqual(self.work_request.children.count(), 4 * 3 * 2)

    def test_combination_logic_multiple_targets(self) -> None:
        # Add two variations just on the backend
        self.add_target(backends=[BackendType.UNSHARE])
        self.add_target(backends=[BackendType.INCUS_LXC])
        # Add three variations just on the variant
        self.add_target(variants=[None, "lintian"])
        self.add_target(variants=["sbuild"])
        self.add_target(variants=["variant1", "variant2"])
        # We should have 5 work requests
        self.assertEqual(self.work_request.children.count(), 5)

    def test_combination_logic_duplicate_between_targets(self) -> None:
        # Each target represents 4 combinations
        self.add_target(
            codenames=["bookworm", "trixie"], architectures=["amd64", "i386"]
        )
        self.add_target(
            codenames=["trixie", "sid"], architectures=["i386", "s390x"]
        )
        # But 1 of them (trixie/i386) is common so 7 unique environments
        self.assertEqual(self.work_request.children.count(), 7)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Models used for artifact data."""

import abc
import enum
import urllib.parse
from typing import Annotated, Any, Literal, Self, TypeAlias, cast, override

import pydantic
import pyparsing

from debusine.assets import KeyPurpose
from debusine.utils import DjangoChoicesEnum
from debusine.utils.tag_utils import DerivationRuleParser


class TaskTypes(DjangoChoicesEnum):
    """Possible values for task_type."""

    WORKER = "Worker"
    SERVER = "Server"
    INTERNAL = "Internal"
    WORKFLOW = "Workflow"
    SIGNING = "Signing"
    WAIT = "Wait"


class WorkRequestResults(DjangoChoicesEnum):
    """Possible values for WorkRequest.result."""

    NONE = ""
    SUCCESS = "success"
    FAILURE = "failure"
    ERROR = "error"
    SKIPPED = "skipped"

    @override
    @enum.property
    def label(self) -> str:
        return self.capitalize()


class BaseArtifactDataModel(pydantic.BaseModel):
    """Base pydantic model for bare/artifact data and their components."""

    model_config = pydantic.ConfigDict(validate_assignment=True, extra="forbid")

    # Replace with serialize_by_alias=True in model_config once we can
    # assume pydantic >= 2.11.
    @override
    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        """Use aliases by default when serializing."""
        kwargs.setdefault("by_alias", True)
        return super().model_dump(**kwargs)


class BareDataCategory(enum.StrEnum):
    """Possible categories for bare data in collection items."""

    PACKAGE_BUILD_LOG = "debian:package-build-log"
    PROMISE = "debusine:promise"
    HISTORICAL_TASK_RUN = "debusine:historical-task-run"
    TASK_CONFIGURATION = "debusine:task-configuration"
    QA_RESULT = "debian:qa-result"

    # Only implemented in tests.
    TEST = "debusine:test"


class DebusinePromise(BaseArtifactDataModel):
    """Pydantic model for debusine promise data."""

    promise_work_request_id: int
    promise_workflow_id: int
    promise_category: str

    # Allow extra fields.
    model_config = BaseArtifactDataModel.model_config | pydantic.ConfigDict(
        extra="allow"
    )

    @pydantic.model_validator(mode="after")
    def check_no_extra_names_starting_promise(self) -> Self:
        r"""Forbid extra field names starting with 'promise\\_'."""
        for key in self.model_extra or {}:
            if key.startswith("promise_"):
                raise ValueError(
                    f"Field name '{key}' starting with "
                    f"'promise_' is not allowed."
                )

        return self


class RuntimeStatistics(BaseArtifactDataModel):
    """Runtime statistics about a completed task."""

    #: Runtime duration of task, in seconds.
    duration: int | None = None

    #: Amount of CPU time used (both user and system), in seconds.
    cpu_time: int | None = None

    #: Maximum disk space used, in bytes.
    disk_space: int | None = None

    #: Maximum RAM used, in bytes.
    memory: int | None = None

    #: Available disk space when the task started, in bytes; may be rounded.
    available_disk_space: int | None = None

    #: Available RAM when the task started, in bytes; may be rounded.
    available_memory: int | None = None

    #: Number of CPU cores on the worker that ran the task.
    cpu_count: int | None = None


def _validate_no_colons(value: str | None) -> str | None:
    """
    Check that value, if set, contains no colons.

    This is used, for example, when colons in a value would break the framing
    of the computed CollectionItem name.
    """
    if value is not None and ":" in value:
        raise ValueError(f"{value!r} may not contain ':'")
    return value


class DebusineHistoricalTaskRun(BaseArtifactDataModel):
    """Pydantic model for historical task runs."""

    task_type: TaskTypes
    task_name: str
    subject: str | None = None
    context: str | None = None
    timestamp: int
    work_request_id: int
    result: WorkRequestResults
    runtime_statistics: RuntimeStatistics


def safe_yaml_relpath(value: str | None) -> str | None:
    """Validate a valid relative path."""
    if value is None:
        return None
    if not value.endswith((".yaml", ".yml")):
        raise ValueError(f"{value!r} must end with .yaml or .yml")
    if value.startswith("/"):
        raise ValueError(f"{value!r} must not be an absolute path")
    components = value.split("/")
    if "." in components:
        raise ValueError(f"{value!r} must not contain '.' as a path element")
    if ".." in components:
        raise ValueError(f"{value!r} must not contain '..' as a path element")
    return value


_derivation_rule_parser = DerivationRuleParser()


def taglist_or_tagexpr(value: str | list[str]) -> str | list[str]:
    """Validate a tag list or tag expression."""
    if isinstance(value, str):
        # Parse the expression to validate it
        try:
            _derivation_rule_parser.as_expr(value)
        except pyparsing.ParseException as e:
            raise ValueError("invalid tag expression: " + str(e))
    return value


class DebusineTaskConfiguration(BaseArtifactDataModel):
    """Pydantic model for task configuration."""

    # Note: this is used in debusine.client.models to deserialize the contents
    # of a task configuration collection. If this model gets changed, the old
    # version needs to be copied to debusine.client.models, and
    # debusine.server.serializers.TaskConfigurationCollectionContentsSerializer
    # changed accordingly, to preserve compatibility with old clients

    task_type: TaskTypes | None = None
    task_name: Annotated[
        str | None, pydantic.AfterValidator(_validate_no_colons)
    ] = None
    subject: Annotated[
        str | None, pydantic.AfterValidator(_validate_no_colons)
    ] = None
    context: str | None = None
    path: Annotated[str | None, pydantic.AfterValidator(safe_yaml_relpath)] = (
        None
    )
    position: int | None = None
    tags_provided: Annotated[
        str | list[str], pydantic.AfterValidator(taglist_or_tagexpr)
    ] = []
    tags_required: Annotated[
        str | list[str], pydantic.AfterValidator(taglist_or_tagexpr)
    ] = []
    template: Annotated[
        str | None, pydantic.AfterValidator(_validate_no_colons)
    ] = None

    use_templates: list[str] = []
    delete_values: list[str] = []
    default_values: dict[str, Any] = {}
    override_values: dict[str, Any] = {}
    lock_values: list[str] = []
    provide_tags: list[str] = []
    require_tags: list[str] = []
    comment: str = ""

    def name(self) -> str:
        """Return the bare data item name for this task configuration."""
        if self.template is not None:
            return f"template:{self.template}"
        elif self.path is not None:
            return f"{self.path}:{self.position or 0}"
        else:
            context = urllib.parse.quote(self.context) if self.context else ''
            return (
                f"{self.task_type}:{self.task_name}"
                f":{self.subject or ''}:{context}"
            )

    @staticmethod
    def get_lookup_names(
        task_type: str,
        task_name: str,
        subject: str | None,
        context: str | None,
    ) -> list[str]:
        """Compute names to look up to get a task's configuration items."""
        names: list[str] = []
        context_encoded = urllib.parse.quote(context) if context else ''
        # Global (subject=None, context=None)
        names.append(f"{task_type}:{task_name}::")
        # Context level (subject=None, context != None)
        if context_encoded:
            names.append(f"{task_type}:{task_name}::{context_encoded}")
        # Subject level (subject != None, context=None)
        if subject:
            names.append(f"{task_type}:{task_name}:{subject}:")
        # Specific-combination level (subject != None, context != None)
        if subject and context_encoded:
            names.append(f"{task_type}:{task_name}:{subject}:{context_encoded}")
        return names

    @pydantic.model_validator(mode="after")
    def template_or_task_match(self) -> Self:
        """Ensure we have either template or the other 4 match fields."""
        if self.template:
            present: list[str] = [
                name
                for name in ("task_type", "task_name", "subject", "context")
                if getattr(self, name)
            ]
            if present:
                raise ValueError(
                    f"{', '.join(present)} should be empty if template is set"
                )
        else:
            missing: list[str] = [
                name
                for name in ("task_type", "task_name")
                if not getattr(self, name)
            ]
            if missing:
                raise ValueError(
                    f"{', '.join(missing)} should be set for non-template"
                    " task configuration entries"
                )
        return self

    @pydantic.model_validator(mode="after")
    def path_position_length(self) -> Self:
        """Ensure that path and position can be collection item names."""
        candidate_name = f"{self.path}:{self.position}"
        if len(candidate_name) > 255:
            raise ValueError(
                "path:position string {candidate_name!r}"
                f" is {len(candidate_name)} characters and must be"
                " no longer than 255"
            )
        return self


class DebianQAResult(BaseArtifactDataModel):
    """Pydantic model for a single QA result."""

    task_name: str
    package: str
    version: str
    architecture: str
    timestamp: int
    work_request_id: int
    result: WorkRequestResults


class ArtifactCategory(enum.StrEnum):
    """Possible artifact categories."""

    WORK_REQUEST_DEBUG_LOGS = "debusine:work-request-debug-logs"
    PACKAGE_BUILD_LOG = "debian:package-build-log"
    UPLOAD = "debian:upload"
    SOURCE_PACKAGE = "debian:source-package"
    BINARY_PACKAGE = "debian:binary-package"
    BINARY_PACKAGES = "debian:binary-packages"
    DEBUG_SYMBOLS = "debian:debug-symbols"
    LINTIAN = "debian:lintian"
    AUTOPKGTEST = "debian:autopkgtest"
    PIUPARTS = "debian:piuparts"
    SYSTEM_TARBALL = "debian:system-tarball"
    BLHC = "debian:blhc"
    DEBDIFF = "debian:debdiff"
    SYSTEM_IMAGE = "debian:system-image"
    SIGNING_INPUT = "debusine:signing-input"
    SIGNING_OUTPUT = "debusine:signing-output"
    REPOSITORY_INDEX = "debian:repository-index"

    # Only implemented in tests.
    TEST = "debusine:test"


class CollectionCategory(enum.StrEnum):
    """Possible collection categories."""

    ENVIRONMENTS = "debian:environments"
    SUITE = "debian:suite"
    PACKAGE_BUILD_LOGS = "debian:package-build-logs"
    WORKFLOW_INTERNAL = "debusine:workflow-internal"
    TASK_CONFIGURATION = "debusine:task-configuration"
    TASK_HISTORY = "debusine:task-history"
    ARCHIVE = "debian:archive"
    QA_RESULTS = "debian:qa-results"

    # Only implemented in tests.
    TEST = "debusine:test"


SINGLETON_COLLECTION_CATEGORIES = {
    CollectionCategory.ARCHIVE,
    CollectionCategory.PACKAGE_BUILD_LOGS,
    CollectionCategory.TASK_HISTORY,
}


class ArtifactData(BaseArtifactDataModel, abc.ABC):
    """
    Base class for artifact data.

    Artifact data is encoded as JSON in the database and in the API, and it is
    modeled as a pydantic data structure in memory for both ease of access and
    validation.
    """

    @abc.abstractmethod
    def get_label(self) -> str | None:
        """
        Return a short human-readable label for the artifact.

        :return: None if no label could be computed from artifact data
        """


class EmptyArtifactData(ArtifactData):
    """Placeholder type for artifacts that have empty data."""

    @override
    def get_label(self) -> None:
        """Return a short human-readable label for the artifact."""
        return None


class DoseDistCheckHyphenize(BaseArtifactDataModel):
    """Fix dose3 field convention that uses Python-incompatible hyphens."""

    # Replace all _ by - in fields.
    model_config = BaseArtifactDataModel.model_config | pydantic.ConfigDict(
        alias_generator=lambda x: x.replace("_", "-")
    )


class DoseDistCheckBasePkg(BaseArtifactDataModel):
    """Data for dose-distcheck package, common fields."""

    package: str
    version: str
    architecture: str | None = None


class DoseDistCheckDepchainPkg(DoseDistCheckBasePkg):
    """Data for dose-distcheck dependency chain package."""

    depends: str | None = None


class DoseDistCheckDepchain(BaseArtifactDataModel):
    """Data for dose-distcheck dependency chain."""

    depchain: list[DoseDistCheckDepchainPkg]


class DoseDistCheckReasonMissingPkg(
    DoseDistCheckBasePkg, DoseDistCheckHyphenize
):
    """Data for dose-distcheck reason, missing variant, missing package."""

    unsat_dependency: str | None = None


class DoseDistCheckReasonMissingExt(BaseArtifactDataModel):
    """Actual data for dose-distcheck reason, missing variant."""

    pkg: DoseDistCheckReasonMissingPkg
    depchains: list[DoseDistCheckDepchain] = []


class DoseDistCheckReasonMissing(BaseArtifactDataModel):
    """Data for dose-distcheck reason, missing variant."""

    missing: DoseDistCheckReasonMissingExt


class DoseDistCheckReasonConflictPkg(
    DoseDistCheckBasePkg, DoseDistCheckHyphenize
):
    """Data for dose-distcheck reason, conflict variant, conflicting package."""

    unsat_conflict: str | None = None


class DoseDistCheckReasonConflictExt(BaseArtifactDataModel):
    """Actual data for dose-distcheck reason, conflict variant."""

    pkg1: DoseDistCheckReasonConflictPkg
    pkg2: DoseDistCheckReasonConflictPkg
    depchain1: list[DoseDistCheckDepchain] = []
    depchain2: list[DoseDistCheckDepchain] = []


class DoseDistCheckReasonConflict(BaseArtifactDataModel):
    """Data for dose-distcheck reason, conflict variant."""

    conflict: DoseDistCheckReasonConflictExt


class DoseDistCheckPackage(DoseDistCheckBasePkg):
    """Data for dose-distcheck main package."""

    essential: bool | None = None
    type: str | None = None
    source: str | None = None
    status: Literal["broken"] | Literal["ok"]
    success: str | None = None
    reasons: list[DoseDistCheckReasonMissing | DoseDistCheckReasonConflict]


class DoseDistCheck(DoseDistCheckHyphenize):
    """Dose3 output, limited to dose-debcheck as invoked by sbuild."""

    # original spec (<1.0):
    # https://gitlab.com/irill/dose3/-/blob/master/doc/debcheck/proposals/distcheck.yaml
    # proposed fix: https://gitlab.com/irill/dose3/-/issues/19

    output_version: str
    native_architecture: str | None = None
    foreign_architecture: str | None = None
    host_architecture: str | None = None

    report: list[DoseDistCheckPackage]

    total_packages: int
    broken_packages: int
    background_packages: int
    foreground_packages: int


class DebianPackageBuildLog(ArtifactData):
    """Data for debian:package-build-log artifacts."""

    source: str
    version: str
    architecture: str | None = None
    filename: str
    bd_uninstallable: DoseDistCheck | None = None

    @override
    def get_label(self) -> str:
        """Return a short human-readable label for the artifact."""
        return self.filename.removesuffix(".build")


class DebianUpload(ArtifactData):
    """Data for debian:upload artifacts."""

    type: Literal["dpkg"]
    changes_fields: dict[str, Any]

    @override
    def get_label(self) -> str | None:
        """Return a short human-readable label for the artifact."""
        source = self.changes_fields.get("Source")
        version = self.changes_fields.get("Version")
        if source and version:
            return f"{source}_{version}"
        if files := self.changes_fields.get("Files"):
            for file in files:
                if (name := file["name"]).endswith(".changes"):
                    break
            else:
                name = files[0]["name"]
            assert isinstance(name, str)
            return name
        return None

    @staticmethod
    def _changes_architectures(
        changes_fields: dict[str, Any],
    ) -> frozenset[str]:
        return frozenset(changes_fields["Architecture"].split())

    @staticmethod
    def _changes_filenames(changes_fields: dict[str, Any]) -> list[str]:
        return [file["name"] for file in changes_fields["Files"]]

    @pydantic.field_validator("changes_fields", mode="after")
    @classmethod
    def metadata_mandatory_fields(cls, data: dict[str, Any]) -> dict[str, Any]:
        """Validate that changes_fields contains Architecture and Files."""
        if "Architecture" not in data:
            raise ValueError("changes_fields must contain Architecture")
        if "Files" not in data:
            raise ValueError("changes_fields must contain Files")
        return data

    @pydantic.field_validator("changes_fields")
    @classmethod
    def metadata_contains_debs_if_binary(
        cls, data: dict[str, Any]
    ) -> dict[str, Any]:
        """
        Validate that binary uploads reference binaries.

        And that source uploads don't contain any.
        """
        archs = cls._changes_architectures(data)

        filenames = cls._changes_filenames(data)
        binaries = [
            file for file in filenames if file.endswith((".deb", ".udeb"))
        ]

        if archs == frozenset({"source"}) and binaries:
            raise ValueError(
                f"Unexpected binary packages {binaries} found in source-only "
                f"upload."
            )
        elif archs - frozenset({"source"}) and not binaries:
            raise ValueError(
                f"No .debs found in {sorted(filenames)} which is expected to "
                f"contain binaries for {', '.join(archs)}"
            )
        return data

    @pydantic.field_validator("changes_fields")
    @classmethod
    def metadata_contains_dsc_if_source(
        cls, data: dict[str, Any]
    ) -> dict[str, Any]:
        """
        Validate that source uploads contain one and only one source package.

        And that binary-only uploads don't contain any.
        """
        archs = cls._changes_architectures(data)

        filenames = cls._changes_filenames(data)
        sources = [file for file in filenames if file.endswith(".dsc")]
        archs = cls._changes_architectures(data)

        if "source" in archs and len(sources) != 1:
            raise ValueError(
                f"Expected to find one and only one source package in source "
                f"upload. Found {sources}."
            )
        elif "source" not in archs and sources:
            raise ValueError(
                f"Binary uploads cannot contain source packages. "
                f"Found: {sources}."
            )
        return data


class DebianSourcePackage(ArtifactData):
    """Data for debian:source-package artifacts."""

    name: str
    version: str
    type: Literal["dpkg"]
    dsc_fields: dict[str, Any]

    @override
    def get_label(self) -> str:
        """Return a short human-readable label for the artifact."""
        return f"{self.name}_{self.version}"


class DebianBinaryPackage(ArtifactData):
    """Data for debian:binary-package artifacts."""

    srcpkg_name: str
    srcpkg_version: str
    deb_fields: dict[str, Any]
    deb_control_files: list[str]

    @override
    def get_label(self) -> str:
        """Return a short human-readable label for the artifact."""
        return "_".join(
            (
                self.deb_fields["Package"],
                self.deb_fields["Version"],
                self.deb_fields["Architecture"],
            )
        )


class DebianBinaryPackages(ArtifactData):
    """Data for debian:binary-packages artifacts."""

    srcpkg_name: str
    srcpkg_version: str
    version: str
    architecture: str
    packages: list[str]

    @override
    def get_label(self) -> str:
        """Return a short human-readable label for the artifact."""
        return f"{self.srcpkg_name}_{self.srcpkg_version}"


class DebianDebugSymbols(ArtifactData):
    """Data for debian:debug-symbols artifacts.

    Each instance represents all .debug ELF files extracted from a single
    -dbgsym.deb package. Files are stored in FileInArtifact with
    path = bare 40-char build-ID.
    """

    build_ids: list[Annotated[str, pydantic.Field(pattern=r"^[0-9a-f]{40}$")]]

    @pydantic.field_validator("build_ids")
    @classmethod
    def _no_duplicates(cls, v: list[str]) -> list[str]:
        seen = set()
        for build_id in v:
            if build_id in seen:
                raise ValueError(
                    f"duplicate build_id {build_id!r} in build_ids"
                )
            seen.add(build_id)
        return v

    @override
    def get_label(self) -> str:
        """Return a short human-readable label for the artifact."""
        count = len(self.build_ids)
        return f"debug-symbols ({count} build-id{'s' if count != 1 else ''})"


class DebianSystemTarball(ArtifactData):
    """Data for debian:system-tarball artifacts."""

    filename: str
    vendor: str
    codename: str
    mirror: pydantic.AnyUrl
    components: list[str]
    variant: str | None = None
    pkglist: dict[str, str]
    architecture: str
    with_dev: bool
    with_init: bool

    @override
    def get_label(self) -> str:
        """Return a short human-readable label for the artifact."""
        return self.filename


class DebianSystemImage(DebianSystemTarball):
    """Data for debian:system-image artifacts."""

    image_format: Literal["raw", "qcow2"]
    filesystem: str
    size: int
    boot_mechanism: Literal["efi", "bios"]


class DebianLintianSeverity(enum.StrEnum):
    """Possible values for lintian tag severities."""

    ERROR = "error"
    WARNING = "warning"
    INFO = "info"
    PEDANTIC = "pedantic"
    EXPERIMENTAL = "experimental"
    OVERRIDDEN = "overridden"
    CLASSIFICATION = "classification"


class DebianLintianSummary(BaseArtifactDataModel):
    """Summary of lintian results."""

    tags_count_by_severity: dict[DebianLintianSeverity, int]
    package_filename: dict[str, str]
    tags_found: list[str]
    overridden_tags_found: list[str]
    lintian_version: str
    distribution: str


class DebianLintian(ArtifactData):
    """Data for debian:lintian artifacts."""

    # TODO: Once old artifacts created without it have expired, this should
    # become required.  (It isn't possible to set it in a data migration in
    # all cases.)
    architecture: str | None = None
    summary: DebianLintianSummary

    @override
    def get_label(self) -> str:
        """Return a short human-readable label for the artifact."""
        if self.summary.package_filename:
            return (
                f"lintian: {', '.join(sorted(self.summary.package_filename))}"
            )
        else:
            return "lintian (empty)"


class DebianAutopkgtestResultStatus(enum.StrEnum):
    """Possible values for status."""

    PASS = "PASS"
    FAIL = "FAIL"
    FLAKY = "FLAKY"
    SKIP = "SKIP"


class DebianAutopkgtestResult(BaseArtifactDataModel):
    """A single result for an autopkgtest test."""

    status: DebianAutopkgtestResultStatus
    details: str | None = None


class DebianAutopkgtestSource(BaseArtifactDataModel):
    """The source package for an autopkgtest run."""

    name: str
    version: str
    url: pydantic.AnyUrl


class DebianAutopkgtest(ArtifactData):
    """Data for debian:autopkgtest artifacts."""

    results: dict[str, DebianAutopkgtestResult]
    cmdline: str
    source_package: DebianAutopkgtestSource
    architecture: str
    distribution: str

    @override
    def get_label(self) -> str:
        """Return a short human-readable label for the artifact."""
        return self.source_package.name


class DebianPiuparts(ArtifactData):
    """Data for debian:piuparts artifacts."""

    piuparts_version: str
    binary_packages: dict[str, str]
    architecture: str
    distribution: str

    @override
    def get_label(self) -> str:
        """Return a short human-readable label for the artifact."""
        return f"piuparts: {', '.join(sorted(self.binary_packages))}"


class DebusineSigningInput(ArtifactData):
    """Input to a Sign task."""

    trusted_certs: list[str] | None = None
    binary_package_name: str | None = None

    @override
    def get_label(self) -> None:
        """Return a short human-readable label for the artifact."""
        return None


class SigningResult(BaseArtifactDataModel):
    """Result of signing a single file."""

    file: str
    output_file: str | None = None
    error_message: str | None = None

    @pydantic.model_validator(mode="after")
    def check_exactly_one_result(self) -> Self:
        """Ensure exactly one of output_file and error_message is present."""
        if (self.output_file is None) == (self.error_message is None):
            raise ValueError(
                "Exactly one of output_file and error_message must be set"
            )
        return self


class DebusineSigningOutput(ArtifactData):
    """Output of a Sign task."""

    purpose: KeyPurpose
    fingerprint: str
    results: list[SigningResult]
    binary_package_name: str | None = None

    @override
    def get_label(self) -> None:
        """Return a short human-readable label for the artifact."""
        return None


class DebianBlhc(ArtifactData):
    """Data for debian:blhc artifacts."""

    # TODO: Once old artifacts created without them have expired, these
    # should become required.
    source: str | None = None
    version: str | None = None
    architecture: str | None = None

    @override
    def get_label(self) -> str:
        """Return a short human-readable label for the artifact."""
        if self.architecture is None:
            return "blhc analysis"
        else:
            return (
                f"blhc analysis for"
                f" {self.source}_{self.version}_{self.architecture}"
            )


class DebDiff(ArtifactData):
    """Data for debian:debdiff artifacts."""

    original: str
    new: str
    label: str | None = None

    @override
    def get_label(self) -> str:
        """Return a short human-readable label for the artifact."""
        if self.label is not None:
            return self.label

        return f"debdiff {self.original} {self.new}"


class DebianRepositoryIndex(ArtifactData):
    """Data for debian:repository-index artifacts."""

    # New artifacts have this set; old ones may not.
    path: str | None = None

    @override
    def get_label(self) -> str | None:
        """Return a short human-readable label for the artifact."""
        return self.path


#: Index ArtifactData model classes by category
ARTIFACT_DATA_MODELS_BY_CATEGORY: dict[ArtifactCategory, type[ArtifactData]] = {
    ArtifactCategory.AUTOPKGTEST: DebianAutopkgtest,
    ArtifactCategory.BINARY_PACKAGE: DebianBinaryPackage,
    ArtifactCategory.BINARY_PACKAGES: DebianBinaryPackages,
    ArtifactCategory.BLHC: DebianBlhc,
    ArtifactCategory.DEBDIFF: DebDiff,
    ArtifactCategory.DEBUG_SYMBOLS: DebianDebugSymbols,
    ArtifactCategory.LINTIAN: DebianLintian,
    ArtifactCategory.PACKAGE_BUILD_LOG: DebianPackageBuildLog,
    ArtifactCategory.PIUPARTS: DebianPiuparts,
    ArtifactCategory.REPOSITORY_INDEX: DebianRepositoryIndex,
    ArtifactCategory.SIGNING_INPUT: DebusineSigningInput,
    ArtifactCategory.SIGNING_OUTPUT: DebusineSigningOutput,
    ArtifactCategory.SOURCE_PACKAGE: DebianSourcePackage,
    ArtifactCategory.SYSTEM_IMAGE: DebianSystemImage,
    ArtifactCategory.SYSTEM_TARBALL: DebianSystemTarball,
    ArtifactCategory.TEST: EmptyArtifactData,
    ArtifactCategory.UPLOAD: DebianUpload,
    ArtifactCategory.WORK_REQUEST_DEBUG_LOGS: EmptyArtifactData,
}


def get_source_package_name(
    artifact_data: (
        DebianSourcePackage
        | DebianUpload
        | DebianBinaryPackage
        | DebianBinaryPackages
        | DebianPackageBuildLog
    ),
) -> str:
    """Retrieve the source name of a package."""
    match artifact_data:
        case DebianSourcePackage():
            return artifact_data.name
        case DebianUpload():
            return cast(str, artifact_data.changes_fields["Source"])
        case DebianBinaryPackage() | DebianBinaryPackages():
            return artifact_data.srcpkg_name
        case DebianPackageBuildLog():
            return artifact_data.source

    raise TypeError(f"Unexpected type: {type(artifact_data).__name__}")


def get_source_package_version(
    artifact_data: (
        DebianSourcePackage
        | DebianBinaryPackage
        | DebianBinaryPackages
        | DebianUpload
        | DebianPackageBuildLog
    ),
) -> str:
    """Retrieve the source version of a package."""
    match artifact_data:
        case DebianSourcePackage():
            return artifact_data.version
        case DebianBinaryPackage() | DebianBinaryPackages():
            return artifact_data.srcpkg_version
        case DebianUpload():
            return cast(str, artifact_data.changes_fields["Version"])
        case DebianPackageBuildLog():
            return artifact_data.version

    raise TypeError(f"Unexpected type: {type(artifact_data).__name__}")


def get_binary_package_name(artifact_data: DebianBinaryPackage) -> str:
    """Retrieve the binary package name."""
    match artifact_data:
        case DebianBinaryPackage():
            return cast(str, artifact_data.deb_fields.get("Package"))

    raise TypeError(f"Unexpected type: {type(artifact_data).__name__}")


def get_binary_package_version(
    artifact_data: DebianUpload | DebianBinaryPackage,
) -> str:
    """Retrieve the binary package version."""
    match artifact_data:
        case DebianUpload():
            return cast(str, artifact_data.changes_fields["Version"])
        case DebianBinaryPackage():
            return cast(str, artifact_data.deb_fields["Version"])

    raise TypeError(f"Unexpected type: {type(artifact_data).__name__}")


ArtifactDataWithArch: TypeAlias = (
    DebianBinaryPackages
    | DebianBinaryPackage
    | DebianPackageBuildLog
    | DebianUpload
)


def get_architectures(artifact_data: ArtifactDataWithArch) -> set[str]:
    """Retrieve the architecture of the package."""
    architectures: set[str] = set()
    match artifact_data:
        case DebianBinaryPackages() | DebianPackageBuildLog():
            assert artifact_data.architecture is not None
            architectures.add(artifact_data.architecture)
        case DebianBinaryPackage():
            architectures.add(artifact_data.deb_fields["Architecture"])
        case DebianUpload():
            for architecture in artifact_data.changes_fields[
                "Architecture"
            ].split():
                if architecture != "source":
                    architectures.add(architecture)
        case _:
            raise TypeError(f"Unexpected type: {type(artifact_data).__name__}")
    return architectures
